import numpy as np
import torch
from numba import njit

@njit
def _Omega_QUADRATIC_CPU(theta_flat, j_idx, k_idx, values, sigma=1.0):
    """
    Optimized CPU implementation of the quadratic potential using Numba.

    Parameters:
        theta_flat (np.ndarray): shape (J,)
        j_idx (np.ndarray): indices j (shape N_edges,)
        k_idx (np.ndarray): indices k (shape N_edges,)
        values (np.ndarray): edge weights (shape N_edges,)
        sigma (float): standard deviation (scalar)

    Returns:
        grad_U (np.ndarray): shape (J,)
        hess_U (np.ndarray): shape (J,)
        U_value (float): scalar
    """
    n_nodes = theta_flat.shape[0]
    n_edges = j_idx.shape[0]

    grad_U = np.zeros(n_nodes)
    hess_U = np.zeros(n_nodes)
    U_value = 0.0

    for i in range(n_edges):
        j = j_idx[i]
        k = k_idx[i]
        v = values[i]
        diff = theta_flat[j] - theta_flat[k]

        psi = 0.5 * (diff / sigma) ** 2
        psi *= v
        U_value += psi

        grad = -v * diff / sigma**2
        hess = v / sigma**2

        grad_U[j] += grad
        hess_U[j] += hess

    U_value *= 0.5
    return grad_U, hess_U, U_value

def _Omega_QUADRATIC_GPU(theta_flat, index, values, device, sigma=1.0):
    """
    GPU implementation of the quadratic potential function, gradient and Hessian.
    
    Parameters:
        theta_flat (torch.Tensor): (J,) tensor on GPU
        index (Tuple[torch.Tensor, torch.Tensor]): (j_idx, k_idx), indices of adjacent pixels
        values (torch.Tensor): (N_edges,) weights, typically 1 or distance-based
        sigma (float): smoothness hyperparameter
        
    Returns:
        grad_U (torch.Tensor): gradient of the potential function, shape (J,)
        hess_U (torch.Tensor): diagonal of the Hessian, shape (J,)
        U_value (torch.Tensor): scalar, energy
    """
    j_idx, k_idx = index
    diff = theta_flat[j_idx] - theta_flat[k_idx]

    # Energy
    psi_pair = 0.5 * (diff / sigma) ** 2
    psi_pair = values * psi_pair

    # Gradient
    grad_pair = values * (-diff / sigma**2)

    # Hessian
    hess_pair = values * (1.0 / sigma**2)

    # Allocate buffers on correct device
    grad_U = torch.zeros_like(theta_flat, device=device)
    hess_U = torch.zeros_like(theta_flat, device=device)

    # Accumulate
    grad_U.index_add_(0, j_idx, grad_pair)
    hess_U.index_add_(0, j_idx, hess_pair)

    U_value = 0.5 * psi_pair.sum()

    return grad_U, hess_U, U_value


from enum import Enum

class PhantomType(Enum):
    """
    Enum for different types of acoustic phantoms.

    Selection of phantom types:
    """
    PVA = 'PVA'
    """Polyvinyl Alcohol phantom.""" 
    Homogeneous = 'Homogeneous'   
    """Homogeneous phantom (water or uniform medium)."""


from AOT_biomaps.AOT_Recon.ReconTools import power_method, gradient, div, proj_l2, prox_G, prox_F_star, _call_axpby, _call_minus_axpy, compute_TV_cpu, power_method_estimate_L__SELL, calculate_memory_requirement, check_gpu_memory
from AOT_biomaps.Config import config
from AOT_biomaps.AOT_Recon.ReconEnums import NoiseType, SMatrixType
import torch
from tqdm import trange
import numpy as np
import pycuda.driver as drv

'''
This module implements Primal-Dual Hybrid Gradient (PDHG) methods for solving inverse problems in Acousto-Optic Tomography.
It includes Chambolle-Pock algorithms for Total Variation (TV) and Kullback-Leibler (KL) divergence regularization.
The methods can run on both CPU and GPU, with configurations set in the AOT_biomaps.Config module.
'''

def CP_TV(
    SMatrix, 
    y, 
    alpha=None,               # TV regularization parameter (if None, alpha is auto-scaled)
    beta=1e-4,              # Tikhonov regularization parameter     
    theta=1.0,
    numIterations=5000, 
    isSavingEachIteration=True,
    L=None, 
    withTumor=True,
    device=None,
    max_saves=5000,
    show_logs=True,
    smatrixType=SMatrixType.SELL,
    k_security=0.8,
    use_power_method=True,
    auto_alpha_gamma=0.05,    # gamma for auto alpha: alpha = gamma * data_term / tv_term
    apply_positivity_clamp=True,
    tikhonov_as_gradient=False,  # if True, apply -tau*2*beta*x instead of prox multiplicative
    use_laplacian=True,         # enable Laplacian (Hessian scalar) penalty
    laplacian_beta_scale=1.0 # multiply beta for laplacian term if you want separate scaling
):
    # try:
    tumor_str = "WITH" if withTumor else "WITHOUT"
    # Auto-select device and method
    if device is None:
        if torch.cuda.is_available() and check_gpu_memory(config.select_best_gpu(), calculate_memory_requirement(SMatrix, y), show_logs=show_logs):
            device = torch.device(f"cuda:{config.select_best_gpu()}")
            use_gpu = True
        else:
            device = torch.device("cpu")
            use_gpu = False
    else:
        use_gpu = device.type == "cuda"
    # Dispatch to the appropriate implementation
    if use_gpu:
            if smatrixType == SMatrixType.CSR:
                raise NotImplementedError("GPU Chambolle Pock (LS-TV) with CSR not implemented.")
            elif smatrixType == SMatrixType.SELL:
                return CP_TV_Tikhonov_sparseSELL_pycuda(SMatrix, y, alpha,beta, theta, numIterations, isSavingEachIteration, L, tumor_str, device, max_saves, show_logs, k_security, use_power_method, auto_alpha_gamma, apply_positivity_clamp, tikhonov_as_gradient, use_laplacian, laplacian_beta_scale)
            elif smatrixType == SMatrixType.DENSE:
                return CP_TV_dense(SMatrix, y, alpha, theta, numIterations, isSavingEachIteration, L, tumor_str, device, max_saves, show_logs)
            else:
                raise ValueError("Unsupported SMatrixType for GPU Chambolle Pock (LS-TV).")
    else:
        raise NotImplementedError("CPU Chambolle Pock (LS-TV) not implemented.")

def CP_KL(
    SMatrix, 
    y, 
    alpha=None,               # TV regularization parameter (if None, alpha is auto-scaled)
    beta=1e-4,              # Tikhonov regularization parameter     
    theta=1.0,
    numIterations=5000, 
    isSavingEachIteration=True,
    L=None, 
    withTumor=True,
    device=None,
    max_saves=5000,
    show_logs=True,
    smatrixType=SMatrixType.SELL,
    k_security=0.8,
    use_power_method=True,
    auto_alpha_gamma=0.05,    # gamma for auto alpha: alpha = gamma * data_term / tv_term
    apply_positivity_clamp=True,
    tikhonov_as_gradient=False,  # if True, apply -tau*2*beta*x instead of prox multiplicative
    use_laplacian=True,         # enable Laplacian (Hessian scalar) penalty
    laplacian_beta_scale=1.0 # multiply beta for laplacian term if you want separate scaling
):
    # try:
    tumor_str = "WITH" if withTumor else "WITHOUT"
    # Auto-select device and method
    if device is None:
        if torch.cuda.is_available() and check_gpu_memory(config.select_best_gpu(), calculate_memory_requirement(SMatrix, y), show_logs=show_logs):
            device = torch.device(f"cuda:{config.select_best_gpu()}")
            use_gpu = True
        else:
            device = torch.device("cpu")
            use_gpu = False
    else:
        use_gpu = device.type == "cuda"
    # Dispatch to the appropriate implementation
    if use_gpu:
            if smatrixType == SMatrixType.CSR:
                raise NotImplementedError("GPU Chambolle Pock (LS-KL) with CSR not implemented.")
            elif smatrixType == SMatrixType.SELL:
                raise NotImplementedError("GPU Chambolle Pock (LS-KL) with SELL not implemented.")
            elif smatrixType == SMatrixType.DENSE:
                return CP_KL(SMatrix, y, alpha, theta, numIterations, isSavingEachIteration, L, tumor_str, device, max_saves, show_logs)
            else:
                raise ValueError("Unsupported SMatrixType for GPU Chambolle Pock (LS-KL).")
    else:
        raise NotImplementedError("CPU Chambolle Pock (LS-KL) not implemented.")




def CP_TV_dense(
    SMatrix,
    y,
    alpha=1e-1,
    theta=1.0,
    numIterations=5000,
    isSavingEachIteration=True,
    L=None,
    withTumor=True,
    device=None,
    max_saves=5000,
    show_logs=True,
):
    """
    Chambolle-Pock algorithm for Total Variation (TV) regularization.
    Works on both CPU and GPU.
    Args:
        SMatrix: System matrix (shape: T, Z, X, N)
        y: Measurement data (shape: T, N)
        alpha: Regularization parameter for TV
        theta: Relaxation parameter (1.0 for standard Chambolle-Pock)
        numIterations: Number of iterations
        isSavingEachIteration: If True, returns selected intermediate reconstructions
        L: Lipschitz constant (estimated if None)
        withTumor: Boolean for description only
        device: Torch device (auto-selected if None)
        max_saves: Maximum number of intermediate saves (default: 5000)
    """
    # Auto-select device if not provided
    if device is None:
        device = torch.device(f"cuda:{config.select_best_gpu()}" if torch.cuda.is_available() else "cpu")
    else:
        device = torch.device(device)

    # Convert data to tensors and move to device
    A = torch.tensor(SMatrix, dtype=torch.float32, device=device)
    y = torch.tensor(y, dtype=torch.float32, device=device)
    T, Z, X, N = SMatrix.shape
    A_flat = A.permute(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y.reshape(-1)

    # Robust normalization
    norm_A = A_flat.abs().max().clamp(min=1e-8)
    norm_y = y_flat.abs().max().clamp(min=1e-8)
    A_flat = A_flat / norm_A
    y_flat = y_flat / norm_y

    # Define forward/backward operators
    P = lambda x: torch.matmul(A_flat, x)
    PT = lambda y: torch.matmul(A_flat.T, y)

    # Estimate Lipschitz constant if needed
    if L is None:
        try:
            L = power_method(P, PT, y_flat, Z, X)
            L = max(L, 1e-3)
        except:
            L = 1.0

    sigma = 1.0 / L
    tau = 1.0 / L

    # Initialize variables
    x = torch.zeros(Z * X, device=device)
    p = torch.zeros((2, Z, X), device=device)
    q = torch.zeros_like(y_flat)
    x_tilde = x.clone()

    # Calculate save indices
    if numIterations <= max_saves:
        save_indices = list(range(numIterations))
    else:
        step = numIterations // max_saves
        save_indices = list(range(0, numIterations, step))
        if save_indices[-1] != numIterations - 1:
            save_indices.append(numIterations - 1)

    I_reconMatrix = []
    saved_indices = []

    # Description for progress bar
    tumor_str = "WITH TUMOR" if withTumor else "WITHOUT TUMOR"
    device_str = f"GPU no.{torch.cuda.current_device()}" if device.type == "cuda" else "CPU"
    description = f"AOT-BioMaps -- Primal/Dual Reconstruction (LS-TV) α:{alpha:.4f} L:{L:.4f} -- {tumor_str} -- {device_str}"

    iterator = trange(numIterations, desc=description) if show_logs else range(numIterations)
    for it in iterator:
        # Update p (TV proximal step)
        grad_x = gradient(x_tilde.reshape(Z, X))
        p = proj_l2(p + sigma * grad_x, alpha)

        # Update q (data fidelity)
        q = (q + sigma * (P(x_tilde) - y_flat)) / (1 + sigma)

        # Update x
        x_old = x.clone()
        div_p = div(p).ravel()  # Divergence calculation
        ATq = PT(q)
        x = (x - tau * (ATq - div_p)) / (1 + tau * 1e-6)  # Light L2 regularization

        # Update x_tilde
        x_tilde = x + theta * (x - x_old)

        # Save intermediate result if needed
        if isSavingEachIteration and it in save_indices:
            I_reconMatrix.append(x.reshape(Z, X).clone() * (norm_y / norm_A))
            saved_indices.append(it)

    # Return results
    if isSavingEachIteration:
        return [tensor.cpu().numpy() for tensor in I_reconMatrix], saved_indices
    else:
        return (x.reshape(Z, X) * (norm_y / norm_A)).cpu().numpy(), None

def CP_TV_Tikhonov_sparseSELL_pycuda(
    SMatrix,
    y,
    alpha=None,               # TV regularization parameter (if None, alpha is auto-scaled)
    beta=1e-4,              # Tikhonov regularization parameter     
    theta=1.0,
    numIterations=2000,
    isSavingEachIteration=True,
    L=None,
    tumor_str="",
    device=None,
    max_saves=2000,
    show_logs=True,
    k_security=0.8,
    use_power_method=True,
    auto_alpha_gamma=0.05,    # gamma for auto alpha: alpha = gamma * data_term / tv_term
    apply_positivity_clamp=True,
    tikhonov_as_gradient=False,  # if True, apply -tau*2*beta*x instead of prox multiplicative
    use_laplacian=True,         # enable Laplacian (Hessian scalar) penalty
    laplacian_beta_scale=1.0    # multiply beta for laplacian term if you want separate scaling
):
    
    """
    CP-TV + Tikhonov + Laplacian (Hessian scalar) penalty integrated.
    Returns (I_reconMatrix, saved_indices) if isSavingEachIteration else (x_final, None).
    """
        # ----- begin main -----
    if SMatrix.ctx:
        SMatrix.ctx.push()

    # prepare variables
    dtype = np.float32
    TN = int(SMatrix.N * SMatrix.T)
    ZX = int(SMatrix.Z * SMatrix.X)
    Z, X = SMatrix.Z, SMatrix.X
    block_size = 256

    # existing kernels
    projection_kernel = SMatrix.sparse_mod.get_function("projection_kernel__SELL")
    backprojection_kernel = SMatrix.sparse_mod.get_function("backprojection_kernel__SELL")
    axpby_kernel = SMatrix.sparse_mod.get_function("vector_axpby_kernel")
    minus_axpy_kernel = SMatrix.sparse_mod.get_function("vector_minus_axpy_kernel")
    gradient_kernel = SMatrix.sparse_mod.get_function("gradient_kernel")
    divergence_kernel = SMatrix.sparse_mod.get_function("divergence_kernel")
    proj_tv_kernel = SMatrix.sparse_mod.get_function("proj_tv_kernel")

    # optional kernels (laplacian & clamp)
    has_laplacian = False
    has_clamp_kernel = False
    try:
        laplacian_kernel = SMatrix.sparse_mod.get_function("laplacian_kernel")
        laplacian_adj_kernel = SMatrix.sparse_mod.get_function("laplacian_adj_kernel")
        has_laplacian = True
    except Exception:
        has_laplacian = False

    try:
        clamp_positive_kernel = SMatrix.sparse_mod.get_function("clamp_positive_kernel")
        has_clamp_kernel = True
    except Exception:
        has_clamp_kernel = False

    stream = drv.Stream()

    # estimate L operator norm if needed
    if use_power_method or L is None:
        L_LS_sq = power_method_estimate_L__SELL(SMatrix, stream, n_it=20, block_size=block_size)
        L_nabla_sq = 8.0
        L_op_norm = np.sqrt(L_LS_sq + L_nabla_sq)
        if L_op_norm < 1e-6:
            L_op_norm = 1.0
    else:
        L_op_norm = L

    tau = np.float32(k_security / L_op_norm)
    sigma = np.float32(k_security / L_op_norm)

    # prepare y and normalization
    y = y.T.astype(dtype).reshape(-1)
    maxy = float(np.max(np.abs(y))) if y.size > 0 else 0.0
    if maxy > 0:
        y_normed = (y / maxy).copy()
    else:
        y_normed = y.copy()

    # GPU allocations
    bufs = []
    y_gpu = drv.mem_alloc(y_normed.nbytes); bufs.append(y_gpu)
    drv.memcpy_htod_async(y_gpu, y_normed.T.flatten(), stream)

    x_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(x_gpu)
    drv.memset_d32_async(x_gpu, 0, ZX, stream)
    x_old_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(x_old_gpu)
    x_tilde_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(x_tilde_gpu)
    drv.memcpy_dtod_async(x_tilde_gpu, x_gpu, ZX * np.dtype(dtype).itemsize, stream)

    p_gpu = drv.mem_alloc(2 * ZX * np.dtype(dtype).itemsize); bufs.append(p_gpu)
    q_gpu = drv.mem_alloc(TN * np.dtype(dtype).itemsize); bufs.append(q_gpu)
    drv.memset_d32_async(p_gpu, 0, 2 * ZX, stream)
    drv.memset_d32_async(q_gpu, 0, TN, stream)

    grad_gpu = drv.mem_alloc(2 * ZX * np.dtype(dtype).itemsize); bufs.append(grad_gpu)
    div_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(div_gpu)
    Ax_gpu = drv.mem_alloc(TN * np.dtype(dtype).itemsize); bufs.append(Ax_gpu)
    ATq_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(ATq_gpu)
    zero_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(zero_gpu)
    drv.memset_d32_async(zero_gpu, 0, ZX, stream)

    # Laplacian buffers (if enabled and kernel available)
    use_lap = use_laplacian and has_laplacian and (beta > 0)
    if use_lap:
        lap_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(lap_gpu)
        r_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(r_gpu)
        drv.memset_d32_async(r_gpu, 0, ZX, stream)
        # scalar beta for laplacian (allow separate scale)
        beta_lap = float(beta) * float(laplacian_beta_scale)
        inv_1_plus_sigma_beta = np.float32(1.0 / (1.0 + float(sigma) * beta_lap))

    # host buffers for logs
    x_host = np.empty(ZX, dtype=dtype)
    Ax_host = np.empty(TN, dtype=dtype)
    q_host = np.empty(TN, dtype=dtype)
    p_host = np.empty(2 * ZX, dtype=dtype)
    ATq_host = np.empty(ZX, dtype=dtype)

    # compute initial backprojection for auto-alpha
    drv.memset_d32_async(ATq_gpu, 0, ZX, stream)
    backprojection_kernel(SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
                        y_gpu, ATq_gpu, np.int32(TN), np.int32(SMatrix.slice_height),
                        block=(block_size, 1, 1), grid=((TN + block_size - 1) // block_size, 1, 1), stream=stream)
    stream.synchronize()
    drv.memcpy_dtoh(x_host, ATq_gpu)

    # auto alpha if requested
    if alpha is None:
        drv.memcpy_htod_async(x_gpu, x_host, stream)
        projection_kernel(Ax_gpu, SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
                        x_gpu, np.int32(TN), np.int32(SMatrix.slice_height),
                        block=(block_size, 1, 1), grid=((TN + block_size - 1) // block_size, 1, 1), stream=stream)
        stream.synchronize()
        drv.memcpy_dtoh(Ax_host, Ax_gpu)
        resid = Ax_host - y_normed[:TN]
        data_term = 0.5 * float(np.dot(resid, resid))
        tv_term = float(compute_TV_cpu(x_host, Z, X)) + 1e-12
        alpha = float(auto_alpha_gamma * data_term / tv_term)
        if show_logs:
            print(f"[auto-alpha] data_term={data_term:.6e}, tv_term={tv_term:.6e}, alpha_set={alpha:.6e}")

    # tikhonov prox multiplicative scale
    if tikhonov_as_gradient:
        tikh_scale = None
    else:
        tikh_scale = np.float32(1.0 / (1.0 + 2.0 * tau * beta)) if beta > 0 else np.float32(1.0)

    # saving policy
    if numIterations <= max_saves:
        save_indices_all = list(range(0, numIterations + 1))
    else:
        step = max(1, numIterations // max_saves)
        save_indices_all = list(range(0, numIterations + 1, step))

    device_str = f"GPU no.{torch.cuda.current_device()}" if device.type == "cuda" else "CPU"
    if show_logs:
        if (alpha is None or alpha == 0) and (beta is None or beta == 0):
            print(f"Parameters: L={L_op_norm:.6e} tau={tau:.3e} sigma={sigma:.3e} lap_enabled={use_lap}")
            description = f"AOT-BioMaps -- Primal/Dual Reconstruction (LS) -- {tumor_str} -- {device_str}"
        if alpha is None or alpha == 0:
            print(f"Parameters: L={L_op_norm:.6e} tau={tau:.3e} sigma={sigma:.3e} beta={beta:.4e} lap_enabled={use_lap}")
            description = f"AOT-BioMaps -- Primal/Dual Reconstruction (LS-Tikhonov) -- {tumor_str} -- {device_str}"
        elif beta is None or beta == 0:
            print(f"Parameters: L={L_op_norm:.6e} tau={tau:.3e} sigma={sigma:.3e} alpha={alpha:.4e} beta={beta:.4e} lap_enabled={use_lap}")
            description = f"AOT-BioMaps -- Primal/Dual Reconstruction (LS-TV) -- {tumor_str} -- {device_str}"
        else:
            print(f"Parameters: L={L_op_norm:.6e} tau={tau:.3e} sigma={sigma:.3e} alpha={alpha:.4e} beta={beta:.4e} lap_enabled={use_lap}")
            description = f"AOT-BioMaps -- Primal/Dual Reconstruction (LS-TV-Tikhonov) -- {tumor_str} -- {device_str}"

    I_reconMatrix = []
    saved_indices = []
    if isSavingEachIteration and 0 in save_indices_all:
        drv.memcpy_dtoh(x_host, x_gpu)
        x0 = x_host.reshape((Z, X)).copy()
        if maxy > 0:
            x0 *= maxy
        I_reconMatrix.append(x0)
        saved_indices.append(0)

    # main loop
    try:
        iterator = trange(numIterations, desc=description) if show_logs else range(numIterations)
        for it in iterator:
            # 1) dual p update (TV)
            gradient_kernel(grad_gpu, x_tilde_gpu, np.int32(Z), np.int32(X), np.int32(ZX),
                            block=(block_size, 1, 1),
                            grid=((X + block_size - 1) // block_size, (Z + block_size - 1) // block_size, 1),
                            stream=stream)
            _call_axpby(axpby_kernel, p_gpu, p_gpu, grad_gpu, 1.0, sigma, 2 * ZX, stream, block_size)
            proj_tv_kernel(p_gpu, np.float32(alpha), np.int32(ZX),
                            block=(block_size, 1, 1),
                            grid=((ZX + block_size - 1) // block_size, 1, 1),
                            stream=stream)

            # 2) dual q update (data fidelity)
            projection_kernel(Ax_gpu, SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
                            x_tilde_gpu, np.int32(TN), np.int32(SMatrix.slice_height),
                            block=(block_size, 1, 1), grid=((TN + block_size - 1) // block_size, 1, 1), stream=stream)
            _call_axpby(axpby_kernel, Ax_gpu, Ax_gpu, y_gpu, 1.0, -1.0, TN, stream, block_size)
            _call_axpby(axpby_kernel, q_gpu, q_gpu, Ax_gpu, 1.0 / (1.0 + sigma), sigma / (1.0 + sigma), TN, stream, block_size)

            # optional Laplacian dual update
            if use_lap:
                # compute Laplacian of x_tilde -> lap_gpu
                laplacian_kernel(lap_gpu, x_tilde_gpu, np.int32(Z), np.int32(X), np.int32(ZX),
                                block=(block_size, 1, 1),
                                grid=((X + block_size - 1) // block_size, (Z + block_size - 1) // block_size, 1),
                                stream=stream)
                # r = r + sigma * lap
                _call_axpby(axpby_kernel, r_gpu, r_gpu, lap_gpu, 1.0, sigma, ZX, stream, block_size)
                # r = r / (1 + sigma * beta_lap)
                _call_axpby(axpby_kernel, r_gpu, r_gpu, zero_gpu, inv_1_plus_sigma_beta, 0.0, ZX, stream, block_size)

            # 3) primal x update
            drv.memcpy_dtod_async(x_old_gpu, x_gpu, ZX * np.dtype(dtype).itemsize, stream)
            divergence_kernel(div_gpu, p_gpu, np.int32(Z), np.int32(X), np.int32(ZX),
                            block=(block_size, 1, 1),
                            grid=((X + block_size - 1) // block_size, (Z + block_size - 1) // block_size, 1),
                            stream=stream)
            drv.memset_d32_async(ATq_gpu, 0, ZX, stream)
            backprojection_kernel(SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
                                q_gpu, ATq_gpu, np.int32(TN), np.int32(SMatrix.slice_height),
                                block=(block_size, 1, 1), grid=((TN + block_size - 1) // block_size, 1, 1), stream=stream)
            # ATq - div
            _call_minus_axpy(minus_axpy_kernel, ATq_gpu, div_gpu, 1.0, ZX, stream, block_size)

            # if laplacian is used, add H^T r into ATq
            if use_lap:
                # compute laplacian_adj_kernel(temp, r)
                # reuse grad_gpu as temporary if safe (its content used earlier, but not reused until later)
                laplacian_adj_kernel(grad_gpu, r_gpu, np.int32(Z), np.int32(X), np.int32(ZX),
                                    block=(block_size, 1, 1),
                                    grid=((X + block_size - 1) // block_size, (Z + block_size - 1) // block_size, 1),
                                    stream=stream)
                # ATq_gpu += temp (grad_gpu)
                _call_axpby(axpby_kernel, ATq_gpu, ATq_gpu, grad_gpu, 1.0, 1.0, ZX, stream, block_size)

            # x = x_old - tau * ATq_buffer
            _call_minus_axpy(minus_axpy_kernel, x_gpu, ATq_gpu, tau, ZX, stream, block_size)

            # Tikhonov
            if beta > 0:
                if tikhonov_as_gradient:
                    mul = 1.0 - 2.0 * float(tau) * float(beta)
                    if mul <= 0.0:
                        # fallback to prox multiplicative stable
                        fallback_scale = np.float32(1.0 / (1.0 + 2.0 * float(tau) * float(beta)))
                        _call_axpby(axpby_kernel, x_gpu, x_gpu, zero_gpu, fallback_scale, 0.0, ZX, stream, block_size)
                    else:
                        # x *= mul  => implemented as axpby: out = 1* x + (mul-1)*x
                        _call_axpby(axpby_kernel, x_gpu, x_gpu, x_gpu, 1.0, np.float32(mul - 1.0), ZX, stream, block_size)
                else:
                    _call_axpby(axpby_kernel, x_gpu, x_gpu, zero_gpu, tikh_scale, np.float32(0.0), ZX, stream, block_size)

            # positivity clamp (prefer GPU kernel if available)
            if apply_positivity_clamp:
                if has_clamp_kernel:
                    # in-place clamp on GPU
                    clamp_positive_kernel(x_gpu, np.int32(ZX),
                                        block=(block_size, 1, 1),
                                        grid=((ZX + block_size - 1) // block_size, 1, 1),
                                        stream=stream)
                else:
                    # fallback CPU roundtrip (slower)
                    stream.synchronize()
                    drv.memcpy_dtoh(x_host, x_gpu)
                    np.maximum(x_host, 0.0, out=x_host)
                    drv.memcpy_htod_async(x_gpu, x_host, stream)

            # extrapolation
            _call_axpby(axpby_kernel, x_tilde_gpu, x_gpu, x_old_gpu, np.float32(1.0 + theta), np.float32(-theta), ZX, stream, block_size)

            # saves
            if isSavingEachIteration and (it + 1) in save_indices_all:
                stream.synchronize()
                drv.memcpy_dtoh(x_host, x_gpu)
                x_saved = x_host.reshape((Z, X)).copy()
                if maxy > 0:
                    x_saved *= maxy
                I_reconMatrix.append(x_saved)
                saved_indices.append(it + 1)

        stream.synchronize()
        drv.memcpy_dtoh(x_host, x_gpu)
        x_final = x_host.reshape((Z, X)).copy()
        if maxy > 0:
            x_final *= maxy
            if isSavingEachIteration and len(I_reconMatrix):
                for i in range(len(I_reconMatrix)):
                    I_reconMatrix[i] *= maxy

        # free buffers
        for buff in bufs:
            try:
                buff.free()
            except:
                pass

        if SMatrix.ctx:
            SMatrix.ctx.pop()

        if isSavingEachIteration:
            return I_reconMatrix, saved_indices
        else:
            return x_final, None

    except Exception as e:
        # cleanup robustly
        print("Error in CP_TV_Tikhonov+Lap (robust):", e)
        try:
            for buff in bufs:
                try:
                    buff.free()
                except:
                    pass
        except:
            pass
        try:
            if SMatrix and hasattr(SMatrix, 'ctx') and SMatrix.ctx:
                SMatrix.ctx.pop()
        except:
            pass
        raise

def CP_KL(
    SMatrix,
    y,
    alpha=1e-9,
    theta=1.0,
    numIterations=5000,
    isSavingEachIteration=True,
    L=None,
    tumor_str="",
    device=None,
    max_saves=5000,
    show_logs=True,
):
    """
    Chambolle-Pock algorithm for Kullback-Leibler (KL) divergence regularization.
    Works on both CPU and GPU.
    Args:
        SMatrix: System matrix (shape: T, Z, X, N)
        y: Measurement data (shape: T, X, N)
        alpha: Regularization parameter
        theta: Relaxation parameter (1.0 for standard Chambolle-Pock)
        numIterations: Number of iterations
        isSavingEachIteration: If True, returns selected intermediate reconstructions
        L: Lipschitz constant (estimated if None)
        withTumor: Boolean for description only
        device: Torch device (auto-selected if None)
        max_saves: Maximum number of intermediate saves (default: 5000)
    """
    # Auto-select device if not provided
    if device is None:
        device = torch.device(f"cuda:{config.select_best_gpu()}" if torch.cuda.is_available() else "cpu")
    else:
        device = torch.device(device)

    # Convert data to tensors and move to device
    A = torch.tensor(SMatrix, dtype=torch.float32, device=device)
    y = torch.tensor(y, dtype=torch.float32, device=device)
    T, Z, X, N = SMatrix.shape
    A_flat = A.permute(0, 3, 1, 2).reshape(T * N, Z * X)
    y_flat = y.reshape(-1)

    # Define forward/backward operators
    P = lambda x: torch.matmul(A_flat, x.ravel())
    PT = lambda y: torch.matmul(A_flat.T, y)

    # Estimate Lipschitz constant if needed
    if L is None:
        L = power_method(P, PT, y_flat, Z, X)

    sigma = 1.0 / L
    tau = 1.0 / L

    # Initialize variables
    x = torch.zeros(Z * X, device=device)
    q = torch.zeros_like(y_flat)
    x_tilde = x.clone()

    # Calculate save indices
    if numIterations <= max_saves:
        save_indices = list(range(numIterations))
    else:
        step = numIterations // max_saves
        save_indices = list(range(0, numIterations, step))
        if save_indices[-1] != numIterations - 1:
            save_indices.append(numIterations - 1)

    I_reconMatrix = [x.reshape(Z, X).cpu().numpy()]
    saved_indices = [0]

    # Description for progress bar
    device_str = f"GPU no.{torch.cuda.current_device()}" if device.type == "cuda" else "CPU"
    description = f"AOT-BioMaps -- Primal/Dual Reconstruction (KL) α:{alpha:.4f} L:{L:.4f} -- {tumor_str} -- {device_str}"

    iterator = trange(numIterations, desc=description) if show_logs else range(numIterations)
    for iteration in iterator:
        # Update q (proximal step for F*)
        q = prox_F_star(q + sigma * P(x_tilde) - sigma * y_flat, sigma, y_flat)

        # Update x (proximal step for G)
        x_old = x.clone()
        x = prox_G(x - tau * PT(q), tau, PT(torch.ones_like(y_flat)))

        # Update x_tilde
        x_tilde = x + theta * (x - x_old)

        # Save intermediate result if needed
        if isSavingEachIteration and iteration in save_indices:
            I_reconMatrix.append(x.reshape(Z, X).cpu().numpy())
            saved_indices.append(iteration)

    # Return results
    if isSavingEachIteration:
        return I_reconMatrix, saved_indices
    else:
        return I_reconMatrix[-1], None
    
def CP_TV_Tikhonov_Preconditioned_SELL_pycuda(
    SMatrix,
    y,
    alpha=None,             # TV param
    beta=1e-4,              # Tikhonov param
    numIterations=2000,
    isSavingEachIteration=True,
    tumor_str="",
    device=None,
    max_saves=2000,
    show_logs=True,
    k_security=1.0,         # Pock condition: tau*sigma * ||A||^2 < 1. Avec precond, souvent k=1.0 suffit.
    apply_positivity_clamp=True
):
    
    if SMatrix.ctx:
        SMatrix.ctx.push()

    dtype = np.float32
    TN = int(SMatrix.N * SMatrix.T)
    ZX = int(SMatrix.Z * SMatrix.X)
    Z, X = SMatrix.Z, SMatrix.X
    block_size = 256

    # --- Récupération des Kernels (Existants + Nouveaux) ---
    projection_kernel = SMatrix.sparse_mod.get_function("projection_kernel__SELL")
    backprojection_kernel = SMatrix.sparse_mod.get_function("backprojection_kernel__SELL")
    axpby_kernel = SMatrix.sparse_mod.get_function("vector_axpby_kernel")
    minus_axpy_kernel = SMatrix.sparse_mod.get_function("vector_minus_axpy_kernel")
    gradient_kernel = SMatrix.sparse_mod.get_function("gradient_kernel")
    divergence_kernel = SMatrix.sparse_mod.get_function("divergence_kernel")
    proj_tv_kernel = SMatrix.sparse_mod.get_function("proj_tv_kernel")
    
    # Nouveaux kernels pour le préconditionnement
    try:
        update_dual_data_precond = SMatrix.sparse_mod.get_function("update_dual_data_precond_kernel")
        update_primal_precond = SMatrix.sparse_mod.get_function("update_primal_precond_kernel")
        invert_vector_kernel = SMatrix.sparse_mod.get_function("invert_vector_kernel")
        clamp_positive_kernel = SMatrix.sparse_mod.get_function("clamp_positive_kernel")
    except Exception as e:
        print("Erreur: Il manque les kernels C++ de préconditionnement (update_dual_data_precond_kernel, etc).")
        raise e

    stream = drv.Stream()

    # --- 1. SETUP PRECONDITIONERS (tau_vec, sigma_vec) ---
    # Au lieu de calculer L, on calcule les sommes lignes/colonnes
    
    # Allocations vecteurs de pas
    sigma_vec_gpu = drv.mem_alloc(TN * np.dtype(dtype).itemsize)
    tau_vec_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize)
    
    # Buffer temporaire de 1 pour les sommes
    ones_TN_gpu = drv.mem_alloc(TN * np.dtype(dtype).itemsize)
    ones_ZX_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize)
    
    # Remplir de 1.0
    # On utilise une astuce : axpby(out, out, out, 0, 0) pour clear puis ajout de 1 ? Non, memset n'écrit pas des float 1.0.
    # On va transférer un array de 1 depuis le host (rapide au setup)
    drv.memcpy_htod_async(ones_TN_gpu, np.ones(TN, dtype=dtype), stream)
    drv.memcpy_htod_async(ones_ZX_gpu, np.ones(ZX, dtype=dtype), stream)
    
    # A) Calcul de Sigma (Inverse des Row Sums) : A * 1_ZX
    # On utilise ton kernel de projection existant : q = A * theta (avec theta=1)
    projection_kernel(sigma_vec_gpu, SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
                      ones_ZX_gpu, np.int32(TN), np.int32(SMatrix.slice_height),
                      block=(block_size, 1, 1), grid=((TN + block_size - 1) // block_size, 1, 1), stream=stream)
    
    # Inversion : sigma_vec = 1 / (sum + eps)
    # Note : sigma_vec doit être divisé par k_security ? Dans Pock precond, tau_i = 1/sum. On garde k_security global si besoin.
    invert_vector_kernel(sigma_vec_gpu, sigma_vec_gpu, np.float32(1e-6), np.int32(TN),
                         block=(block_size, 1, 1), grid=((TN + block_size - 1) // block_size, 1, 1), stream=stream)

    # B) Calcul de Tau (Inverse des Col Sums - Carte Sensibilité) : A^T * 1_TN
    # On utilise ton kernel de backprojection : c += A^T * e (avec e=1)
    # Important : backproj fait un atomicAdd, il faut mettre à 0 avant
    drv.memset_d32_async(tau_vec_gpu, 0, ZX, stream)
    backprojection_kernel(SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
                          ones_TN_gpu, tau_vec_gpu, np.int32(TN), np.int32(SMatrix.slice_height),
                          block=(block_size, 1, 1), grid=((TN + block_size - 1) // block_size, 1, 1), stream=stream)
    
    invert_vector_kernel(tau_vec_gpu, tau_vec_gpu, np.float32(1e-6), np.int32(ZX),
                         block=(block_size, 1, 1), grid=((ZX + block_size - 1) // block_size, 1, 1), stream=stream)

    # Nettoyage temp
    ones_TN_gpu.free()
    ones_ZX_gpu.free()

    # --- 2. PREPARE DATA ---
    y = y.T.astype(dtype).reshape(-1)
    maxy = float(np.max(np.abs(y))) if y.size > 0 else 0.0
    y_normed = (y / maxy).copy() if maxy > 0 else y.copy()

    # Allocations Variables Principales
    bufs = []
    y_gpu = drv.mem_alloc(y_normed.nbytes); bufs.append(y_gpu)
    drv.memcpy_htod_async(y_gpu, y_normed, stream)

    x_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(x_gpu)
    drv.memset_d32_async(x_gpu, 0, ZX, stream)
    
    x_old_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(x_old_gpu)
    x_tilde_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(x_tilde_gpu)
    drv.memset_d32_async(x_tilde_gpu, 0, ZX, stream) # Important: x_tilde init à 0

    # Dual Variables
    p_gpu = drv.mem_alloc(2 * ZX * np.dtype(dtype).itemsize); bufs.append(p_gpu) # Gradient
    q_gpu = drv.mem_alloc(TN * np.dtype(dtype).itemsize); bufs.append(q_gpu)     # Data term
    drv.memset_d32_async(p_gpu, 0, 2 * ZX, stream)
    drv.memset_d32_async(q_gpu, 0, TN, stream)

    # Buffers intermédiaires
    grad_gpu = drv.mem_alloc(2 * ZX * np.dtype(dtype).itemsize); bufs.append(grad_gpu)
    div_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(div_gpu)
    Ax_gpu = drv.mem_alloc(TN * np.dtype(dtype).itemsize); bufs.append(Ax_gpu)
    ATq_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize); bufs.append(ATq_gpu)

    # --- PARAMETRES & LOGS ---
    # Pour le TV, on a besoin d'un sigma scalaire pour le dual update du gradient.
    # Dans le preconditionning diagonal complet, c'est complexe.
    # Approximation robuste : on utilise sigma_tv fixe.
    # Cependant, alpha (le poids TV) joue le rôle de seuil.
    sigma_tv = np.float32(1.0) # Le scaling est absorbé par alpha et tau_vec

    if alpha is None:
        # Auto-alpha simplifié (ou tu gardes ta logique précédente)
        alpha = 0.01 # Valeur par défaut si non spécifié
        print(f"Warning: Alpha not set, using default {alpha}")

    # Description
    desc = f"Pock-Precond | {tumor_str} | Alpha={alpha} | Beta={beta}"
    
    # Saving setup
    x_host = np.empty(ZX, dtype=dtype)
    save_indices_all = list(range(0, numIterations + 1, max(1, numIterations // max_saves)))
    I_reconMatrix = []
    saved_indices = []

    try:
        iterator = trange(numIterations, desc=desc) if show_logs else range(numIterations)
        
        for it in iterator:
            
            # =======================
            # 1. UPDATE DUAL (P - TV)
            # =======================
            # Gradient de x_bar (x_tilde)
            gradient_kernel(grad_gpu, x_tilde_gpu, np.int32(Z), np.int32(X), np.int32(ZX),
                            block=(block_size, 1, 1), grid=((X + block_size - 1) // block_size, (Z + block_size - 1) // block_size, 1), stream=stream)
            
            # p = p + sigma_tv * grad
            axpby_kernel(p_gpu, p_gpu, grad_gpu, np.float32(1.0), sigma_tv, np.int32(2 * ZX),
                         block=(block_size, 1, 1), grid=((2*ZX + block_size - 1) // block_size, 1, 1), stream=stream)
            
            # Prox TV (Projection boule unité)
            # Attention : ici le pas "alpha" doit être adapté car on utilise un tau vectoriel ailleurs.
            # On passe alpha brut.
            proj_tv_kernel(p_gpu, np.float32(alpha), np.int32(ZX),
                           block=(block_size, 1, 1), grid=((ZX + block_size - 1) // block_size, 1, 1), stream=stream)

            # ==========================
            # 2. UPDATE DUAL (Q - DATA)
            # ==========================
            # Calcul Ax_bar
            projection_kernel(Ax_gpu, SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
                              x_tilde_gpu, np.int32(TN), np.int32(SMatrix.slice_height),
                              block=(block_size, 1, 1), grid=((TN + block_size - 1) // block_size, 1, 1), stream=stream)
            
            # Update q avec sigma vectoriel
            # q = (q + sigma_vec * (Ax - y)) / (1 + sigma_vec)
            update_dual_data_precond(q_gpu, Ax_gpu, y_gpu, sigma_vec_gpu, np.int32(TN),
                                     block=(block_size, 1, 1), grid=((TN + block_size - 1) // block_size, 1, 1), stream=stream)

            # ==========================
            # 3. UPDATE PRIMAL (X)
            # ==========================
            # Sauvegarde x_old
            drv.memcpy_dtod_async(x_old_gpu, x_gpu, ZX * np.dtype(dtype).itemsize, stream)
            
            # Calcul ATq (Backprojection)
            drv.memset_d32_async(ATq_gpu, 0, ZX, stream)
            backprojection_kernel(SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
                                  q_gpu, ATq_gpu, np.int32(TN), np.int32(SMatrix.slice_height),
                                  block=(block_size, 1, 1), grid=((TN + block_size - 1) // block_size, 1, 1), stream=stream)
            
            # Calcul Div p
            divergence_kernel(div_gpu, p_gpu, np.int32(Z), np.int32(X), np.int32(ZX),
                              block=(block_size, 1, 1), grid=((X + block_size - 1) // block_size, (Z + block_size - 1) // block_size, 1), stream=stream)
            
            # Combiner les gradients : ATq - Div p
            minus_axpy_kernel(ATq_gpu, div_gpu, np.float32(1.0), np.int32(ZX),
                             block=(block_size, 1, 1), grid=((ZX + block_size - 1) // block_size, 1, 1), stream=stream)

            # Update X avec tau vectoriel
            # x = x - tau_vec * (ATq - Div p)
            # Note: ATq_gpu contient maintenant le gradient total
            update_primal_precond(x_gpu, ATq_gpu, tau_vec_gpu, np.int32(ZX),
                                  block=(block_size, 1, 1), grid=((ZX + block_size - 1) // block_size, 1, 1), stream=stream)

            # Positivity
            if apply_positivity_clamp:
                clamp_positive_kernel(x_gpu, np.int32(ZX),
                                      block=(block_size, 1, 1), grid=((ZX + block_size - 1) // block_size, 1, 1), stream=stream)

            # Tikhonov (simplifié : decay)
            if beta > 0:
                # x = x / (1 + beta * tau_vec) ? 
                # Approx rapide : x *= (1 - beta * mean_tau) ou juste x *= (1-beta)
                # Utilisons axpby simple pour un petit decay
                axpby_kernel(x_gpu, x_gpu, x_gpu, np.float32(1.0 - beta), np.float32(0.0), np.int32(ZX),
                             block=(block_size, 1, 1), grid=((ZX + block_size - 1) // block_size, 1, 1), stream=stream)

            # ==========================
            # 4. EXTRAPOLATION
            # ==========================
            # x_tilde = x + theta * (x - x_old)
            axpby_kernel(x_tilde_gpu, x_gpu, x_old_gpu, np.float32(2.0), np.float32(-1.0), np.int32(ZX),
                         block=(block_size, 1, 1), grid=((ZX + block_size - 1) // block_size, 1, 1), stream=stream)

            # Save
            if isSavingEachIteration and (it + 1) in save_indices_all:
                stream.synchronize()
                drv.memcpy_dtoh(x_host, x_gpu)
                x_saved = x_host.reshape((Z, X)).copy()
                if maxy > 0: x_saved *= maxy
                I_reconMatrix.append(x_saved)
                saved_indices.append(it + 1)

        # Fin boucle
        stream.synchronize()
        # Cleanup (buffers)
        for b in bufs: b.free()
        sigma_vec_gpu.free()
        tau_vec_gpu.free()
        if SMatrix.ctx: SMatrix.ctx.pop()

        return I_reconMatrix, saved_indices

    except Exception as e:
        print(f"Error: {e}")
        if SMatrix.ctx: SMatrix.ctx.pop()
        return None, None
    



from abc import abstractmethod
import os
from joblib import dump, load
from kwave.kgrid import kWaveGrid
from math import ceil
import numpy as np
import matplotlib.pyplot as plt

class Medium:    
    
    def __init__(self, params):
        self.params = params
        self.medium = None
        self.factorX = None
        self.factorZ = None
        self.factorT = None
        self.c_mean = None
        self.Nx_reshaped = None
        self.Nz_reshaped = None
        self.dx_reshaped = None

        self.kgrid = kWaveGrid([self.params.general["Nx"], self.params.general["Nz"]], [self.params.general["dx"], self.params.general["dz"]])

        if self.params.acoustic['f_AQ'] == None:
            self.kgrid.makeTime(self.params.acoustic['medium']['c0'])
            self.params.acoustic['f_AQ'] = int(1/self.kgrid.dt)
        else:
            if self.params.general['Nt'] is None:
                Nt = ceil((self.params.general['Zrange'][1] - self.params.general['Zrange'][0])*float(params.acoustic['f_AQ']) / self.params.acoustic['c0'])
                self.params.general['Nt'] = Nt
            else:
                Nt = self.params.general['Nt']
            self.kgrid.setTime(Nt,1/float(self.params.acoustic['f_AQ']))


    @abstractmethod
    def generate_medium(self):
        """
        Abstract method to generate the medium properties.
        This method should be implemented by subclasses.
        """
        pass

    def save_medium(self, folderPath, fileName = "medium"):
        """
        Save the medium properties to a .joblib file.

        Parameters:
        - folderPath (str): The directory where the .joblib file will be saved.
        - fileName (str): The name of the .joblib file (without extension).
        """
        try:
            os.makedirs(folderPath, exist_ok=True)
            filePath = os.path.join(folderPath, fileName)
            # check if filename ends with an extension
            if not os.path.splitext(fileName)[1]:
                filePath += '.joblib'
            else:
                raise ValueError("The fileName should not contain an extension; .joblib will be added automatically.")
            dump(self, filePath)
        except Exception as e:
            print(f"Error in save_medium method: {e}")
            raise
    
    def load_medium(self, folderPath, fileName = "medium"):
        """
        Load the medium properties from a .joblib file.

        Parameters:
        - folderPath (str): The directory where the .joblib file will be loaded from.
        - fileName (str): The name of the .joblib file (without extension).
        """
        try:          
            if not os.path.splitext(fileName)[1]:
                fileName += '.joblib'
            else:
                raise ValueError("The fileName should not contain an extension; .joblib will be added automatically.")
            filePath = os.path.join(folderPath, fileName)
            if os.path.exists(filePath) == False:
                raise FileNotFoundError(f"The file {filePath} does not exist.")
            loaded_obj = load(filePath)
            self.kmedium = loaded_obj.kmedium
            self.factorX = loaded_obj.factorX
            self.factorZ = loaded_obj.factorZ
            self.factorT = loaded_obj.factorT
            self.c_mean = loaded_obj.c_mean
            self.Nx_reshaped = loaded_obj.Nx_reshaped
            self.Nz_reshaped = loaded_obj.Nz_reshaped
            self.dx_reshaped = loaded_obj.dx_reshaped
            self.kgrid = loaded_obj.kgrid
        except Exception as e:
            print(f"Error in load_medium method: {e}")
            raise   

    def plot_medium_properties(self):
        if self.kmedium is None:
            raise ValueError("Medium properties are not available. Please generate or load the medium first.")
        vmin_speed = np.min(self.kmedium.sound_speed)
        vmax_speed = np.max(self.kmedium.sound_speed)
        vmin_density = np.min(self.kmedium.density)
        vmax_density = np.max(self.kmedium.density)
        extent = [self.params.general['Xrange'][0]*1e3, self.params.general['Xrange'][1]*1e3, self.params.general['Zrange'][1]*1e3, self.params.general['Zrange'][0]*1e3]
        plt.figure(figsize=(13,6))
        plt.subplot(121)
        plt.imshow(self.kmedium.sound_speed.T, vmin=vmin_speed, vmax=vmax_speed, cmap='autumn', extent=extent)
        plt.title('Sound speed map (m/s)')
        plt.xlabel('X (mm)')
        plt.ylabel('Z (mm)')
        plt.colorbar()
        plt.subplot(122)
        plt.imshow(self.kmedium.density.T, vmin=vmin_density, vmax=vmax_density, cmap='summer', extent=extent)
        plt.title('Density map (kg/m^3)')
        plt.xlabel('X (mm)')
        plt.ylabel('Z (mm)')
        plt.colorbar()
        plt.tight_layout()
        plt.show()



import copy
import AOT_biomaps.Settings
from AOT_biomaps.Config import config
from AOT_biomaps.AOT_Acoustic.AcousticTools import calculate_envelope_squared, loadmat
from .AcousticTools import next_power_of_2, reshape_field
from .AcousticEnums import TypeSim, Dim, FormatSave, WaveType
from AOT_biomaps.AOT_Medium import Medium

from IPython.display import HTML
import h5py
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from kwave.utils.signals import tone_burst
from kwave.ksource import kSource
from kwave.ksensor import kSensor
from kwave.kspaceFirstOrder3D import kspaceFirstOrder3D
from kwave.kspaceFirstOrder2D import kspaceFirstOrder2D
from kwave.options.simulation_options import SimulationOptions
from kwave.options.simulation_execution_options import SimulationExecutionOptions
from AOT_biomaps.Settings import Params


from tempfile import gettempdir
from abc import ABC, abstractmethod
import logging



####### ABSTRACT CLASS #######

class AcousticField(ABC):
    """
    Abstract class to generate and manipulate acoustic fields for ultrasound imaging.
    Provides methods to initialize parameters, generate fields, save and load data, and calculate envelopes.

    Principal parameters:
    - field: Acoustic field data.
    - burst: Burst signal used for generating the field for each piezo elements.
    - delayedSignal: Delayed burst signal for each piezo element.
    - medium: Medium properties for k-Wave simulation. Because field2 and Hydrophone simulation are not implemented yet, this attribute is set to None for these types of simulation.
    """

    def __init__(self, params, medium):
        """
        Initialize global properties of the AcousticField object.

        Parameters:
        - typeSim (TypeSim): Type of simulation to be performed. Options include KWAVE, FIELD2, and HYDRO. Default is TypeSim.KWAVE.
        - dim (Dim): Dimension of the acoustic field. Can be 2D or 3D. Default is Dim.D2.
        - c0 (float): Speed of sound in the medium, specified in meters per second (m/s). Default is 1540 m/s.
        - f_US (float): Frequency of the ultrasound signal, specified in Hertz (Hz). Default is 6 MHz.
        - f_AQ (float): Frequency of data acquisition, specified in Hertz (Hz). Default is 180 MHz.
        - f_saving (float): Frequency at which the acoustic field data is saved, specified in Hertz (Hz). Default is 10 MHz.
        - num_cycles (int): Number of cycles in the burst signal. Default is 4 cycles.
        - num_elements (int): Number of elements in the transducer array. Default is 192 elements.
        - element_width (float): Width of each transducer element, specified in meters (m). Default is 0.2 mm.
        - element_height (float): Height of each transducer element, specified in meters (m). Default is 6 mm.
        - Xrange (list of float): Range of X coordinates for the acoustic field, specified in meters (m). Default is from -20 mm to 20 mm.
        - Yrange (list of float, optional): Range of Y coordinates for the acoustic field, specified in meters (m). Default is None, indicating no specific Y range.
        - Zrange (list of float): Range of Z coordinates for the acoustic field, specified in meters (m). Default is from 0 m to 37 mm.
        """
        if type(params) != Params:
            raise TypeError("params must be an instance of the Params class")
        if not isinstance(medium, Medium):
            raise TypeError("medium must be an instance of the Medium class")

        self.medium = medium
        self.params = params
        if self.params.acoustic['typeSim'] != TypeSim.SIMPLE_SIM.value:
            self._generate_burst_signal()
        if self.params.acoustic["dim"] == Dim.D3 and self.params.general["Yrange"] is None:
            raise ValueError("Yrange must be provided for 3D fields.")
            
        self.waveType = None
        self.field = None  

    def __del__(self):
        """
        Destructor for the AcousticField class. Cleans up the field and envelope attributes.
        """
        try:
            self.field = None
            self.burst = None
            self.delayedSignal = None
        except Exception as e:
            print(f"Error in __del__ method: {e}")
            raise

    ## TOOLS METHODS ##

    def generate_field(self, isGpu=config.get_process() == 'gpu',show_log = True):
        """
        Generate the acoustic field based on the specified simulation type and parameters.
        """
        try:
            logging.getLogger('root').setLevel(logging.ERROR)
            if self.params.acoustic['typeSim'] == TypeSim.FIELD2.value:
                raise NotImplementedError("FIELD2 simulation is not implemented yet.")
            elif self.params.acoustic['typeSim'] == TypeSim.SIMPLE_SIM.value:
                self.field = self._generate_acoustic_field_SIMPLE_SIM(show_log)
            elif self.params.acoustic['typeSim'] == TypeSim.KWAVE.value:
                if self.params.acoustic["dim"] == Dim.D2.value:
                    try:
                        field = self._generate_acoustic_field_KWAVE_2D(isGpu, show_log)
                    except Exception as e:
                        raise RuntimeError(f"Failed to generate 2D acoustic field: {e}")
                    self.field = calculate_envelope_squared(field)
                elif self.params.acoustic["dim"] == Dim.D3.value:
                    field = self._generate_acoustic_field_KWAVE_3D(isGpu, show_log)
                    self.field = calculate_envelope_squared(field)
            elif self.params.acoustic['typeSim'] == TypeSim.HYDRO.value:
                raise ValueError("Cannot generate field for Hydrophone simulation, load exciting acquisitions.")
            else:
                raise ValueError("Invalid simulation type. Supported types are: FIELD2, KWAVE, HYDRO.")
        except Exception as e:
            print(f"Error in generate_field method: {e}")
            raise

    def save_field(self, filePath, formatSave=FormatSave.HDR_IMG):
        """
        Save the acoustic field to a file in the specified format.

        Parameters:
        - filePath (str): The path where the file will be saved.
        """
        try:
            if formatSave.value == FormatSave.HDR_IMG.value:
                self._save2D_HDR_IMG(filePath)
            elif formatSave.value == FormatSave.H5.value:
                self._save2D_H5(filePath)
            elif formatSave.value == FormatSave.NPY.value:
                self._save2D_NPY(filePath)
            else:
                raise ValueError("Unsupported format. Supported formats are: HDR_IMG, H5, NPY.")
        except Exception as e:
            print(f"Error in save_field method: {e}")
            raise

    def load_field(self, folderPath, formatSave=FormatSave.HDR_IMG, nameBlock=None):
        """
        Load the acoustic field from a file in the specified format.

        Parameters:
        - filePath (str): The folder path from which to load the file.
        """
        try:
            if str(type(formatSave)) != str(AOT_biomaps.AOT_Acoustic.FormatSave):
                    raise ValueError(f"Unsupported file format: {formatSave}. Supported formats are: HDR_IMG, H5, NPY.")

            if self.params.acoustic['typeSim'] == TypeSim.FIELD2.value:
                raise NotImplementedError("FIELD2 simulation is not implemented yet.")
            elif self.params.acoustic['typeSim'] == TypeSim.KWAVE.value or self.params.acoustic['typeSim'] == TypeSim.SIMPLE_SIM.value:
                if formatSave.value == FormatSave.HDR_IMG.value: 
                    if self.params.acoustic["dim"] == Dim.D2.value:
                        self._load_fieldKWAVE_XZ(os.path.join(folderPath,self.getName_field()+formatSave.value))
                    elif self.params.acoustic["dim"] == Dim.D3.value:
                        raise NotImplementedError("3D KWAVE field loading is not implemented yet.")
                elif formatSave.value == FormatSave.H5.value:
                    if self.params.acoustic["dim"] == Dim.D2.value:
                         self._load_field_h5(folderPath,nameBlock)
                    elif self.params.acoustic["dim"] == Dim.D3.value:
                        raise NotImplementedError("H5 KWAVE field loading is not implemented yet.")
                elif formatSave.value == FormatSave.NPY.value:
                    if self.params.acoustic["dim"] == Dim.D2.value:
                        self.field = np.load(os.path.join(folderPath,self.getName_field()+formatSave.value))
                    elif self.params.acoustic["dim"] == Dim.D3.value:
                        raise NotImplementedError("3D NPY KWAVE field loading is not implemented yet.")
            elif self.params.acoustic['typeSim'] == TypeSim.HYDRO.value:
                print("Loading Hydrophone field...")
                if formatSave.value == FormatSave.HDR_IMG.value:
                    raise ValueError("HDR_IMG format is not supported for Hydrophone acquisition.")
                if formatSave.value == FormatSave.H5.value:
                    if self.params.acoustic["dim"] == Dim.D2.value:
                        self.field, self.params.general['Xrange'], self.params.general['Zrange'] = self._load_fieldHYDRO_XZ(os.path.join(folderPath, self.getName_field() + '.h5'),  os.path.join(folderPath, "PARAMS_" +self.getName_field() + '.mat'))
                    elif self.params.acoustic["dim"] == Dim.D3.value: 
                        self._load_fieldHYDRO_XYZ(os.path.join(folderPath, self.getName_field() + '.h5'),  os.path.join(folderPath, "PARAMS_" +self.getName_field() + '.mat'))
                elif formatSave.value == FormatSave.NPY.value:
                    if self.params.acoustic["dim"] == Dim.D2.value:
                        self.field = np.load(folderPath)
                    elif self.params.acoustic["dim"] == Dim.D3.value:
                        raise NotImplementedError("3D NPY Hydrophone field loading is not implemented yet.")
            else:
                raise ValueError("Invalid simulation type. Supported types are: FIELD2, KWAVE, HYDRO.")
           
        except Exception as e:
            print(f"Error in load_field method: {e}")
            raise

    @abstractmethod
    def getName_field(self):
        pass


    ## DISPLAY METHODS ##

    def plot_burst_signal(self):
        """
        Plot the burst signal used for generating the acoustic field.
        """
        try:
            time2plot = np.arange(0, len(self.burst)) / self.params.acoustic['f_AQ'] * 1000000  # Convert to microseconds
            plt.figure(figsize=(8, 8))
            plt.plot(time2plot, self.burst)
            plt.title('Excitation burst signal')
            plt.xlabel('Time (µs)')
            plt.ylabel('Amplitude')
            plt.grid()
            plt.show()
        except Exception as e:
            print(f"Error in plot_burst_signal method: {e}")
            raise

    def animated_plot_AcousticField(self, desired_duration_ms = 5000, save_dir=None):
        """
        Plot synchronized animations of A_matrix slices for selected angles.

        Args:
            step (int): Time step between frames (default is every 10 frames).
            save_dir (str): Directory to save the animation gif; if None, animation will not be saved.

        Returns:
            ani: Matplotlib FuncAnimation object.
        """
        try:

            maxF = np.max(self.field[:,20:,:])
            minF = np.min(self.field[:,20:,:])
            # Set the maximum embedded animation size to 100 MB
            plt.rcParams['animation.embed_limit'] = 100

            if save_dir is not None:
                os.makedirs(save_dir, exist_ok=True)

            # Create a figure and axis
            fig, ax = plt.subplots()

            # Set main title
            if self.waveType.value == WaveType.FocusedWave.value:
                fig.suptitle("[System Matrix Animation] Focused Wave", fontsize=12, y=0.98)
            elif self.waveType.value == WaveType.PlaneWave.value:
                fig.suptitle(f"[System Matrix Animation] Plane Wave | Angles {self.angle}°", fontsize=12, y=0.98)
            elif self.waveType.value == WaveType.StructuredWave.value:
                fig.suptitle(f"[System Matrix Animation] Structured Wave | Pattern structure: {self.pattern.activeList} | Angles {self.angle}°", fontsize=12, y=0.98)
            else:

                raise ValueError("Invalid wave type. Supported types are: FocusedWave, PlaneWave, StructuredWave.")

            # Initial plot
            im = ax.imshow(
                self.field[0, :, :],
                extent=(self.params.general['Xrange'][0] * 1000, self.params.general['Xrange'][-1] * 1000, self.params.general['Zrange'][-1] * 1000, self.params.general['Zrange'][0] * 1000),
                vmin = 1.2*minF,
                vmax=0.8*maxF,
                aspect='equal',
                cmap='jet',
                animated=True
            )
            ax.set_title(f"t = 0 ms", fontsize=10)
            ax.set_xlabel("x (mm)", fontsize=8)
            ax.set_ylabel("z (mm)", fontsize=8)


            # Unified update function for all subplots
            def update(frame):
                im.set_data(self.field[frame, :, :])
                ax.set_title(f"t = {frame / self.params.acoustic['f_AQ'] * 1000:.2f} ms", fontsize=10)
                return [im]  # Return a list of artists that were modified

            interval = desired_duration_ms / self.field.shape[0]

            # Create animation
            ani = animation.FuncAnimation(
                fig, update,
                frames=range(0, self.field.shape[0]),
                interval=interval, blit=True
            )

            # Save animation if needed
            if save_dir is not None:
                if self.waveType == WaveType.FocusedWave:
                    save_filename = f"Focused_Wave_.gif"
                elif self.waveType == WaveType.PlaneWave:
                    save_filename = f"Plane_Wave_{self._format_angle()}.gif"
                else:
                    save_filename = f"Structured_Wave_PatternStructure_{self.pattern.activeList}_{self._format_angle()}.gif"
                save_path = os.path.join(save_dir, save_filename)
                ani.save(save_path, writer='pillow', fps=20)
                print(f"Saved: {save_path}")

            plt.close(fig)

            return HTML(ani.to_jshtml())
        except Exception as e:
            print(f"Error creating animation: {e}")
            return None

    def show(self, use_dB=False, reference=1e6,Vmax=None):
        """
        Display the maximum intensity projection of the acoustic field envelope.

        Parameters:
        - use_dB (bool): If True, display in dB relative to the reference pressure.
        - reference (float): Reference pressure in Pa for dB calculation (default: 1 MPa).
        """
        try:
            if self.field is None:
                raise ValueError("Field data is not available. Please generate or load the field first.")
            if self.field.min() < 0:
                raise ValueError("Calculation of the envelope has not been performed. Please generate the envelope first.")

            # Convertir l'enveloppe au carré en amplitude (Pa) en prenant la racine carrée
            envelope_amplitude = np.sqrt(self.field)

            if use_dB:
                # Convertir en dB re reference (Pa)
                envelope_dB = 20 * np.log10(envelope_amplitude / reference)
                data_to_show = envelope_dB
                unit_label = f'dB re {reference / 1e6} MPa'
                if Vmax is not None:
                    vmax = Vmax
                else:
                    vmax = 0

            else:
                # Convertir en MPa
                envelope_amplitude_mpa = envelope_amplitude / 1e6
                data_to_show = envelope_amplitude_mpa
                unit_label = 'MPa'
                if Vmax is not None:
                    vmax = Vmax
                else:
                    vmax = 0.85*np.max(envelope_amplitude_mpa)

            plt.figure(figsize=(10, 6))
            plt.imshow(data_to_show.max(axis=0),
                    extent=(self.params.general['Xrange'][0] * 1000, self.params.general['Xrange'][1] * 1000,
                            self.params.general['Zrange'][1] * 1000, self.params.general['Zrange'][0] * 1000),
                    aspect='equal', cmap='jet', vmin=0, vmax=vmax)
            plt.colorbar(label=f'Envelope Amplitude ({unit_label})')
            plt.title('Maximum Intensity Projection of Acoustic Field Envelope')
            plt.xlabel('X (mm)')
            plt.ylabel('Z (mm)')
            plt.show()
        except Exception as e:
            print(f"Error in show method: {e}")
            raise

    ## PRIVATE METHODS ##

    @abstractmethod
    def _generate_acoustic_field_SIMPLE_SIM(self, show_log=False):
        pass

    def _generate_burst_signal(self):
        if self.params.acoustic['typeSim'] == TypeSim.FIELD2.value:
            raise NotImplementedError("FIELD2 simulation is not implemented yet.")
        elif self.params.acoustic['typeSim'] == TypeSim.KWAVE.value:
            self._generate_burst_signalKWAVE()
        elif self.params.acoustic['typeSim'] == TypeSim.HYDRO.value:
            raise ValueError("Cannot generate burst signal for Hydrophone simulation.")

    def _generate_burst_signalKWAVE(self):
        """
        Private method to generate a burst signal based on the specified parameters.
        """
        try:
            self.burst = tone_burst(1/self.medium.kgrid.dt, self.params.acoustic['f_US'], self.params.acoustic['emission']['num_cycles']).squeeze()
        except Exception as e:
            print(f"Error in __generate_burst_signal method: {e}")
            raise

    def _generate_acoustic_field_KWAVE_2D(self, isGPU=True if config.get_process() == 'gpu' else False, show_log=True):
        """
        Base function to generate a 2D acoustic field using k-Wave.
        Handles common setup, simulation, and post-processing.
        """
        source = kSource()
        source.p_mask = np.zeros(( self.medium.Nx_reshaped, self.medium.Nz_reshaped))
        # Appel à la méthode spécialisée
        source = self._SetUpSource(source, self.medium.Nx_reshaped, self.medium.kgrid.dt, self.medium.dx_reshaped, self.medium.c_mean,self.medium.factorT)  # factorT=1 pour simplifier

        # --- 4. Sensor setup ---
        sensor = kSensor()
        sensor.mask = np.ones((self.medium.Nx_reshaped, self.medium.Nz_reshaped))
        # --- 5. PML setup ---
        pml_size = 50 

        # --- 7. Simulation options ---
        simulation_options = SimulationOptions(
        pml_inside=False, # PML ajoutée autour de la grille Air+PVA
        pml_size=[1, pml_size],
        use_sg=False,
        save_to_disk=True,
        input_filename=os.path.join(gettempdir(), "KwaveIN.h5"),
        output_filename=os.path.join(gettempdir(), "KwaveOUT.h5"),
        smooth_c0 = True,
        smooth_rho0 = True,
        smooth_p0 = True
        )

        execution_options = SimulationExecutionOptions(
            is_gpu_simulation=config.get_process() == 'gpu' and isGPU,
            device_num=config.bestGPU,
            show_sim_log=show_log
        )

        medium_copy = copy.deepcopy(self.medium) # Avoid in-place modifications of the medium properties during simulation, which can affect subsequent simulations if the same medium object is reused.

        # --- 8. Run simulation ---
        sensor_data = kspaceFirstOrder2D(
            kgrid=medium_copy.kgrid,
            medium=medium_copy.kmedium,
            source=source,
            sensor=sensor,
            simulation_options=simulation_options,
            execution_options=execution_options,
        )

        # --- 9. Post-process results ---
        data = sensor_data['p'].reshape(self.medium.kgrid.Nt, self.medium.Nz_reshaped, self.medium.Nx_reshaped    )
        if self.medium.factorT != 1 or self.medium.factorX != 1 or self.medium.factorZ != 1:
            data = reshape_field(data, [self.medium.factorT, self.medium.factorX, self.medium.factorZ])
            xStart = (self.medium.Nx_reshaped//2)//self.medium.factorX - (self.params.general['Nx']//2)
            return data[:, :self.params.general['Nx'], xStart:xStart+self.params.general['Nx']]
        else:
            return data[:, :self.params.general['Nx'], xStart:xStart+self.params.general['Nx']]


    # def _generate_acoustic_field_KWAVE_3D(self, isGPU=True, show_log=True):
    #     """
    #     Generate a 3D acoustic field using k-Wave.
    #     """
    #     try:
    #         # --- 1. Grid setup (common) ---
    #         dx = self.params['dx']
    #         if dx >= self.params['element_width']:
    #             dx = self.params['element_width'] / 2
    #             if self.params['width_phantom'] is not None:
    #                 Nx = int(np.round((self.params['width_phantom'])/dx))
    #             else:
    #                 Nx = int(round((self.params['Xrange'][1] - self.params['Xrange'][0]) / dx))
    #             if self.params['height_phantom'] is not None:
    #                 Nz = int(np.round((self.params['height_phantom'])/dx))
    #             else:
    #                 Nz = int(round((self.params['Zrange'][1] - self.params['Zrange'][0]) / dx))
    #         else:
    #             if self.params['width_phantom'] is not None:
    #                 Nx = int(np.round((self.params['width_phantom'])/self.params['dx']))
    #             else:
    #                 Nx = int(round((self.params['Xrange'][1] - self.params['Xrange'][0]) / self.params['dx']))
    #             if self.params['height_phantom'] is not None:
    #                 Nz = int(np.round((self.params['height_phantom'])/self.params['dz']))
    #             else:
    #                 Nz = int(round((self.params['Zrange'][1] - self.params['Zrange'][0]) / self.params['dz']))

    #         # --- 2. Time and space factors (common) ---
    #         factorT = int(np.ceil(self.params['f_AQ'] / self.params['f_saving']))
    #         factorX = int(np.ceil(Nx / self.params['Nx']))
    #         factorZ = int(np.ceil(Nz / self.params['Nz']))

    #         kgrid = kWaveGrid([Nx, Nz], [dx, dx])
    #         kgrid.setTime(self.kgrid.Nt, 1 / self.params['f_AQ'])

    #         source = kSource()
    #         source.p_mask = np.zeros((self.params['Nx'], self.params['Ny'], self.params['Nz']))

    #         # Appel à la méthode spécialisée
    #         self._SetUpSource(source, self.params['Nx'], self.params['dx'], factorT)  # factorT=1 pour simplifier

    #         sensor = kSensor()
    #         sensor.mask = np.ones((self.params['Nx'], self.params['Ny'], self.params['Nz']))

    #         simulation_options = SimulationOptions(
    #             pml_inside=False,
    #             pml_auto=True,
    #             use_sg=False,
    #             save_to_disk=True,
    #             input_filename=os.path.join(gettempdir(), "KwaveIN.h5"),
    #             output_filename=os.path.join(gettempdir(), "KwaveOUT.h5")
    #         )

    #         execution_options = SimulationExecutionOptions(
    #             is_gpu_simulation=config.get_process() == 'gpu' and isGPU,
    #             device_num=config.bestGPU,
    #             show_sim_log=show_log
    #         )

    #         sensor_data = kspaceFirstOrder3D(
    #             kgrid=kgrid,
    #             medium=self.medium,
    #             source=source,
    #             sensor=sensor,
    #             simulation_options=simulation_options,
    #             execution_options=execution_options,
    #         )

    #         data = sensor_data['p'].reshape(kgrid.Nt, Nz, Nx)
    #         if factorT != 1 or factorX != 1 or factorZ != 1:
    #             return reshape_field(data, [factorT, factorX, factorZ])
    #         else:
    #             return data

    #     except Exception as e:
    #         print(f"Error generating 3D acoustic field: {e}")
    #         return None
        
    @abstractmethod
    def _SetUpSource(self, source, Nx, dt, dx, c0, factorT):
        """
        Abstract method: each subclass must implement its own source setup.
        """
        pass

    @abstractmethod
    def _save2D_HDR_IMG(self, filePath):
        """
        Save the 2D acoustic field as an HDR_IMG file.
        Must be implemented in subclasses.
        """
        pass

    def _load_field_h5(self, filePath,nameBlock):
        """
        Load the 2D acoustic field from an H5 file.

        Parameters:
        - filePath (str): The path to the H5 file.

        Returns:
        - field (numpy.ndarray): The loaded acoustic field.
        """
        try:
            if nameBlock is None:
                nameBlock = 'data'
            with h5py.File(os.path.join(filePath, self.getName_field()+".h5"), 'r') as f:
                self.field = f[nameBlock][:]
        except Exception as e:
            print(f"Error in _load_field_h5 method: {e}")
            raise

    def _save2D_H5(self, filePath):
        """
        Save the 2D acoustic field as an H5 file.

        Parameters:
        - filePath (str): The path where the file will be saved.
        """
        try:
            with h5py.File(filePath+self.getName_field()+"h5", 'w') as f:
                for key, value in self.__dict__.items():
                    if key != 'field':
                        f.create_dataset(key, data=value)
                f.create_dataset('data', data=self.field, compression='gzip')
        except Exception as e:
            print(f"Error in _save2D_H5 method: {e}")
            raise

    def _save2D_NPY(self, filePath):
        """
        Save the 2D acoustic field as a NPY file.

        Parameters:
        - filePath (str): The path where the file will be saved.
        """
        try:
            np.save(filePath+self.getName_field()+"npy", self.field)
        except Exception as e:
            print(f"Error in _save2D_NPY method: {e}")
            raise

    def _load_fieldKWAVE_XZ(self, hdr_path):
        """
        Read an Interfile (.hdr) and its binary file (.img) to reconstruct an acoustic field.

        Parameters:
        - hdr_path (str): The path to the .hdr file.

        Returns:
        - field (numpy.ndarray): The reconstructed acoustic field with dimensions reordered to (X, Z, time).
        - header (dict): A dictionary containing the metadata from the .hdr file.
        """
        try:
            header = {}
            # Read the .hdr file
            with open(hdr_path, 'r') as f:
                for line in f:
                    if ':=' in line:
                        key, value = line.split(':=', 1)
                        key = key.strip().lower().replace('!', '')
                        value = value.strip()
                        header[key] = value

            # Get the associated .img file name
            data_file = header.get('name of data file') or header.get('name of date file')
            if data_file is None:
                raise ValueError(f"Cannot find the data file associated with the header file {hdr_path}")
            img_path = os.path.join(os.path.dirname(hdr_path), os.path.basename(data_file))

            # Determine the field size from metadata
            shape = [int(header[f'matrix size [{i}]']) for i in range(1, 3) if f'matrix size [{i}]' in header]
            if not shape:
                raise ValueError("Cannot determine the shape of the acoustic field from metadata.")

            # Data type
            data_type = header.get('number format', 'short float').lower()
            dtype_map = {
                'short float': np.float32,
                'float': np.float32,
                'int16': np.int16,
                'int32': np.int32,
                'uint16': np.uint16,
                'uint8': np.uint8
            }
            dtype = dtype_map.get(data_type)
            if dtype is None:
                raise ValueError(f"Unsupported data type: {data_type}")

            # Byte order (endianness)
            byte_order = header.get('imagedata byte order', 'LITTLEENDIAN').lower()
            endianess = '<' if 'little' in byte_order else '>'

            # Verify the actual size of the .img file
            fileSize = os.path.getsize(img_path)
            timeDim = int(fileSize / (np.dtype(dtype).itemsize * np.prod(shape)))
            shape = shape + [timeDim]

            # Read binary data
            with open(img_path, 'rb') as f:
                data = np.fromfile(f, dtype=endianess + np.dtype(dtype).char)

            # Reshape data to (time, Z, X)
            field = data.reshape(shape[::-1])  # NumPy interprets in C order (opposite of MATLAB)

            # Apply scaling factors if available
            rescale_slope = float(header.get('data rescale slope', 1))
            rescale_offset = float(header.get('data rescale offset', 0))
            field = field * rescale_slope + rescale_offset

            self.field = field
        except Exception as e:
            print(f"Error in _load_fieldKWAVE_XZ method: {e}")
            raise

    def _load_fieldHYDRO_XZ(self, file_path_h5, param_path_mat):
        """
        Load the 2D acoustic field for Hydrophone simulation from H5 and MAT files.

        Parameters:
        - file_path_h5 (str): The path to the H5 file.
        - param_path_mat (str): The path to the MAT file.

        Returns:
        - envelope_transposed (numpy.ndarray): The transposed envelope of the acoustic field.
        """
        try:
            # Load parameters from the .mat file
            param = loadmat(param_path_mat)

            # Load the ranges for x and z
            x_test = param['x'].flatten()
            z_test = param['z'].flatten()

            x_range = np.arange(-23, 21.2, 0.2)
            z_range = np.arange(0, 37.2, 0.2)
            X, Z = np.meshgrid(x_range, z_range)

            # Load the data from the .h5 file
            with h5py.File(file_path_h5, 'r') as file:
                data = file['data'][:]

            # Initialize a matrix to store the acoustic data
            acoustic_field = np.zeros((len(z_range), len(x_range), data.shape[1]))

            # Fill the grid with acoustic data
            index = 0
            for i in range(len(z_range)):
                if i % 2 == 0:
                    # Traverse left to right
                    for j in range(len(x_range)):
                        acoustic_field[i, j, :] = data[index]
                        index += 1
                else:
                    # Traverse right to left
                    for j in range(len(x_range) - 1, -1, -1):
                        acoustic_field[i, j, :] = data[index]
                        index += 1

            # Calculate the analytic envelope
            envelope = np.abs(CPU_hilbert(acoustic_field, axis=2))
            # Reorganize the array to have the shape (Times, Z, X)
            envelope_transposed = np.transpose(envelope, (2, 0, 1)).T

            self.field = envelope_transposed
            self.params.general['Xrange'] = x_range
            self.params.general['Zrange'] = z_range

        except Exception as e:
            print(f"Error in _load_fieldHYDRO_XZ method: {e}")
            raise

    def _load_fieldHYDRO_YZ(self, file_path_h5, param_path_mat):
        """
        Load the 2D acoustic field for Hydrophone simulation from H5 and MAT files.

        Parameters:
        - file_path_h5 (str): The path to the H5 file.
        - param_path_mat (str): The path to the MAT file.

        Returns:
        - envelope_transposed (numpy.ndarray): The transposed envelope of the acoustic field.
        - y_range (numpy.ndarray): The range of y values.
        - z_range (numpy.ndarray): The range of z values.
        """
        try:
            # Load parameters from the .mat file
            param = loadmat(param_path_mat)

            # Extract the ranges for y and z
            y_range = param['y'].flatten()
            z_range = param['z'].flatten()

            # Load the data from the .h5 file
            with h5py.File(file_path_h5, 'r') as file:
                data = file['data'][:]

            # Calculate the number of scans
            Ny = len(y_range)
            Nz = len(z_range)

            # Create the scan positions
            positions_y = []
            positions_z = []

            for i in range(Nz):
                if i % 2 == 0:
                    # Traverse top to bottom for even rows
                    positions_y.extend(y_range)
                else:
                    # Traverse bottom to top for odd rows
                    positions_y.extend(y_range[::-1])
                positions_z.extend([z_range[i]] * Ny)

            Positions = np.column_stack((positions_y, positions_z))

            # Initialize a matrix to store the reorganized data
            reorganized_data = np.zeros((Ny, Nz, data.shape[1]))

            # Reorganize the data according to the scan positions
            for index, (j, k) in enumerate(Positions):
                y_idx = np.where(y_range == j)[0][0]
                z_idx = np.where(z_range == k)[0][0]
                reorganized_data[y_idx, z_idx, :] = data[index, :]

            # Calculate the analytic envelope
            envelope = np.abs(CPU_hilbert(reorganized_data, axis=2))
            # Reorganize the array to have the shape (Times, Z, Y)
            envelope_transposed = np.transpose(envelope, (2, 0, 1))
            return envelope_transposed, y_range, z_range
        except Exception as e:
            print(f"Error in _load_fieldHYDRO_YZ method: {e}")
            raise

    def _load_fieldHYDRO_XYZ(self, file_path_h5, param_path_mat):
        """
        Load the 3D acoustic field for Hydrophone simulation from H5 and MAT files.

        Parameters:
        - file_path_h5 (str): The path to the H5 file.
        - param_path_mat (str): The path to the MAT file.

        Returns:
        - EnveloppeField (numpy.ndarray): The envelope of the acoustic field.
        - x_range (numpy.ndarray): The range of x values.
        - y_range (numpy.ndarray): The range of y values.
        - z_range (numpy.ndarray): The range of z values.
        """
        try:
            # Load parameters from the .mat file
            param = loadmat(param_path_mat)

            # Extract the ranges for x, y, and z
            x_range = param['x'].flatten()
            y_range = param['y'].flatten()
            z_range = param['z'].flatten()

            # Create a meshgrid for x, y, and z
            X, Y, Z = np.meshgrid(x_range, y_range, z_range, indexing='ij')

            # Load the data from the .h5 file
            with h5py.File(file_path_h5, 'r') as file:
                data = file['data'][:]

            # Calculate the number of scans
            Nx = len(x_range)
            Ny = len(y_range)
            Nz = len(z_range)
            Nscans = Nx * Ny * Nz

            # Create the scan positions
            if Ny % 2 == 0:
                X = np.tile(np.concatenate([x_range[:, np.newaxis], x_range[::-1, np.newaxis]]), (Ny // 2, 1))
                Y = np.repeat(y_range, Nx)
            else:
                X = np.concatenate([x_range[:, np.newaxis], np.tile(np.concatenate([x_range[::-1, np.newaxis], x_range[:, np.newaxis]]), ((Ny - 1) // 2, 1))])
                Y = np.repeat(y_range, Nx)

            XY = np.column_stack((X.flatten(), Y))

            if Nz % 2 == 0:
                XYZ = np.tile(np.concatenate([XY, np.flipud(XY)]), (Nz // 2, 1))
                Z = np.repeat(z_range, Nx * Ny)
            else:
                XYZ = np.concatenate([XY, np.tile(np.concatenate([np.flipud(XY), XY]), ((Nz - 1) // 2, 1))])
                Z = np.repeat(z_range, Nx * Ny)

            Positions = np.column_stack((XYZ, Z))

            # Initialize a matrix to store the reorganized data
            reorganized_data = np.zeros((Nx, Ny, Nz, data.shape[1]))

            # Reorganize the data according to the scan positions
            for index, (i, j, k) in enumerate(Positions):
                x_idx = np.where(x_range == i)[0][0]
                y_idx = np.where(y_range == j)[0][0]
                z_idx = np.where(z_range == k)[0][0]
                reorganized_data[x_idx, y_idx, z_idx, :] = data[index, :]

            EnveloppeField = np.zeros_like(reorganized_data)

            for y in range(reorganized_data.shape[1]):
                for z in range(reorganized_data.shape[2]):
                    EnveloppeField[:, y, z, :] = np.abs(CPU_hilbert(reorganized_data[:, y, z, :], axis=1))
            self.field = np.transpose(EnveloppeField,  (3, 2, 1, 0))
            self.params.general['Xrange'] = [x_range[0], x_range[-1]]
            self.params.general['Yrange'] = [y_range[0], y_range[-1]]
            self.params.general['Zrange'] = [z_range[0], z_range[-1]]
            self.params.general['Nx'] = Nx
            self.params.general['Ny'] = Ny
            self.params.general['Nz'] = Nz
        except Exception as e:
            print(f"Error in _load_fieldHYDRO_XYZ method: {e}")
            raise



from ._mainMedium import *
from .HomogeneousMedium import *
from .PVAMedium import *
from .MediumEnums import *

# Docstring for the AOT_Experiment package
"""
AOT_Medium is a package for defining acoustic media in Acousto-Optic Tomography.
It provides classes and enums for different types of media, including homogeneous and PVA phantoms.
"""


import numpy as np
from kwave.kgrid import kWaveGrid
from kwave.kmedium import kWaveMedium
from scipy.ndimage import gaussian_filter
from AOT_biomaps.AOT_Medium._mainMedium import Medium


class PVAMedium(Medium):
    """
    Class representing a homogeneous medium for acoustic wave propagation.
    This class can be used
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def generate_medium(self):
        """
        Génère un milieu k-Wave PVA hétérogène avec gestion optionnelle de l'air.
        - Si isAirReflection=True : l'air entoure le phantom (marges de 20 pixels).
        - Sinon : le PVA occupe toute la grille.
        Respecte Nyquist et optimise la VRAM (float32, calculs in-place).
        """
        # --- 1. Grid setup et respect de Nyquist ---
        dx = self.params.general['dx']
        if dx >= self.params.acoustic['probe']['element_width']:
            dx = self.params.acoustic['probe']['element_width'] / 2
        # Dimensions physiques → pixels
        width = self.params.acoustic['medium'].get('width',
                self.params.general['Xrange'][1] - self.params.general['Xrange'][0])
        height = self.params.acoustic['medium'].get('height',
                self.params.general['Zrange'][1] - self.params.general['Zrange'][0])
        pva_nx = int(np.round(width / dx))
        pva_nz = int(np.round(height / dx))

        # Ajout des marges d'air si nécessaire
        air_margin = 20 if self.params.acoustic['medium']['isAirReflection'] else 0
        Nx, Nz = pva_nx + 2 * air_margin, pva_nz

        # --- 2. space factors ---
        self.factorX = int(np.ceil(self.params.general['dx'] / dx))
        self.factorZ = self.factorX

        # --- 2. Initialisation des cartes (air ou PVA) ---
        c_map = np.full((Nx, Nz), 343.0 if air_margin else self.params.acoustic['medium']['c0'], dtype=np.float32)
        self.c_mean = np.mean(c_map[:, 0])
        rho_map = np.full((Nx, Nz), 1.2 if air_margin else self.params.acoustic['medium']['density'], dtype=np.float32)
        alpha_coeff_map = np.zeros((Nx, Nz), dtype=np.float32)
        BonA_map = np.zeros((Nx, Nz), dtype=np.float32)

        # --- 3. Région du PVA (avec ou sans marge d'air) ---
        x_start = air_margin
        x_end = x_start + pva_nx

        # --- 4. Génération des hétérogènesités (uniquement dans le PVA) ---
        eta = np.random.randn(pva_nx, pva_nz).astype(np.float32) * self.params.acoustic['medium']['noise_lvl']
        for _ in range(self.params.acoustic['medium']['n_phases']):
            sigma_val = np.random.uniform(*[s/self.params.general['dx'] for s in self.params.acoustic['medium']['size_structures']])
            threshold = np.random.uniform(1.2, 2.2)
            noise_field = gaussian_filter(np.random.randn(pva_nx, pva_nz), sigma=sigma_val)
            noise_field /= (np.std(noise_field) + 1e-9)
            mask = 1 / (1 + np.exp(-self.params.acoustic['medium']['grad_coef'] * (noise_field - threshold)))

            zone_offset = np.random.uniform(-self.params.acoustic['medium']['c0_delta'], self.params.acoustic['medium']['c0_delta'])
            zone_grain = np.random.randn(pva_nx, pva_nz).astype(np.float32) * self.params.acoustic['medium']['scattering_amplitude']
            eta = (1 - mask) * eta + mask * (zone_offset + zone_grain)

        # --- 5. Propriétés physiques du PVA ---
        sound_speed_pva = self.params.acoustic['medium']['c0'] * (1 + eta)
        density_pva = self.params.acoustic['medium']['density'] * (1 + eta)
        eta_norm = (eta - np.min(eta)) / (np.max(eta) - np.min(eta) + 1e-9)
        alpha_coeff_pva = 0.4 + 0.3 * eta_norm  # [0.4, 0.7] dB/(MHz^y cm)

        # --- 6. Insertion dans les cartes globales ---
        c_map[x_start:x_end, :] = sound_speed_pva
        rho_map[x_start:x_end, :] = density_pva
        alpha_coeff_map[x_start:x_end, :] = alpha_coeff_pva
        if not air_margin:  # Si pas d'air, BonA/alpha_power uniformes sur toute la grille
            BonA_map[:, :] = self.params.acoustic['medium'].get('BonA', 6.0)
        else:  # Sinon, uniquement dans le PVA
            BonA_map[x_start:x_end, :] = self.params.acoustic['medium'].get('BonA', 6.0)

        c_map = c_map.astype(np.float32)
        rho_map = rho_map.astype(np.float32)
        alpha_coeff_map = alpha_coeff_map.astype(np.float32)
        BonA_map = BonA_map.astype(np.float32)

        # --- 7. Création du milieu k-Wave ---
        self.kmedium = kWaveMedium(
            sound_speed=c_map,
            density=rho_map,
            sound_speed_ref=self.params.acoustic['medium']['c0'],
            alpha_coeff=alpha_coeff_map,
            alpha_power=self.params.acoustic['medium']['alpha_power'],
            BonA=BonA_map,
            absorbing=True,
            stokes=False
        )
               
        self.kgrid = kWaveGrid([Nx, Nz], [dx, dx])
        dt = 1/(self.params.acoustic['f_AQ'])
        self.kgrid.setTime(self.params.general['Nt'], dt)  # Garder Nt constant

        # Saving variable for later use
        self.factorX = int(np.ceil(self.params.general['dx'] / dx))
        self.factorZ = self.factorX
        self.factorT = int(np.ceil((1/self.kgrid.dt) / (self.params.acoustic['f_saving'])))
        self.c_mean = np.mean(c_map[:, 0])
        self.Nx_reshaped = Nx
        self.Nz_reshaped = Nz
        self.dx_reshaped = dx



from enum import Enum

class TypeSim(Enum):
    """
    Enum for the type of simulation to be performed.

    Selection of simulation types:
    - KWAVE: k-Wave simulation.
    - FIELD2: Field II simulation.
    - HYDRO: Hydrophone acquisition.
    """
    KWAVE = 'k-wave'
    """k-Wave simulation."""

    FIELD2 = 'Field2'
    """Field II simulation."""

    HYDRO = 'Hydrophone'
    """Hydrophone acquisition."""

    SIMPLE_SIM = 'SimpleSim'
    """Simple simulation for testing purposes."""

class Dim(Enum):
    """
    Enum for the dimension of the acoustic field.

    Selection of dimensions:
    - D2: 2D field.
    - D3: 3D field.
    """
    D2 = '2D'
    """2D field."""
    D3 = '3D'
    """3D field."""

class FormatSave(Enum):
    """
    Enum for different file formats to save the acoustic field.

    Selection of file formats:
    - HDR_IMG: Interfile format (.hdr and .img).
    - H5: HDF5 format (.h5).
    - NPY: NumPy format (.npy).
    """
    HDR_IMG = '.hdr'
    """Interfile format (.hdr and .img)."""
    H5 = '.h5'
    """HDF5 format (.h5)."""
    NPY = '.npy'
    """NumPy format (.npy)."""

class WaveType(Enum):
    """
    Enum for different types of acoustic waves.

    Selection of wave types:
    - FocusedWave: A wave type where the energy is focused at a specific point.
    - StructuredWave: A wave type characterized by a specific pattern or structure.
    - PlaneWave: A wave type where the wavefronts are parallel and travel in a single direction.
    """
    FocusedWave = 'focus'
    """A wave type where the energy is focused at a specific point."""
    StructuredWave = 'structured'
    """A wave type characterized by a specific pattern or structure."""
    PlaneWave = 'plane'
    """A wave type where the wavefronts are parallel and travel in a single direction."""




import os
from AOT_biomaps.AOT_Recon.AOT_SparseSMatrix import SparseSMatrix_CSR, SparseSMatrix_SELL
import torch
import numpy as np
import pycuda.driver as drv
from numba import njit, prange
from torch_sparse import coalesce
from scipy.signal.windows import hann
from itertools import groupby
import cupy as cp
from cupyx.scipy.ndimage import map_coordinates

def load_recon(hdr_path):
    """
    Lit un fichier Interfile (.hdr) et son fichier binaire (.img) pour reconstruire une image comme le fait Vinci.
    
    Paramètres :
    ------------
    - hdr_path : chemin complet du fichier .hdr
    
    Retour :
    --------
    - image : tableau NumPy contenant l'image
    - header : dictionnaire contenant les métadonnées du fichier .hdr
    """
    header = {}
    with open(hdr_path, 'r') as f:
        for line in f:
            if ':=' in line:
                key, value = line.split(':=', 1)  # s'assurer qu'on ne coupe que la première occurrence de ':='
                key = key.strip().lower().replace('!', '')  # Nettoyage des caractères
                value = value.strip()
                header[key] = value
    
    # 📘 Obtenez le nom du fichier de données associé (le .img)
    data_file = header.get('name of data file')
    if data_file is None:
        raise ValueError(f"Impossible de trouver le fichier de données associé au fichier header {hdr_path}")
    
    img_path = os.path.join(os.path.dirname(hdr_path), data_file)
    
    # 📘 Récupérer la taille de l'image à partir des métadonnées
    shape = [int(header[f'matrix size [{i}]']) for i in range(1, 4) if f'matrix size [{i}]' in header]
    if shape and shape[-1] == 1:  # Si la 3e dimension est 1, on la supprime
        shape = shape[:-1]  # On garde (192, 240) par exemple
    
    if not shape:
        raise ValueError("Impossible de déterminer la forme de l'image à partir des métadonnées.")
    
    # 📘 Déterminez le type de données à utiliser
    data_type = header.get('number format', 'short float').lower()
    dtype_map = {
        'short float': np.float32,
        'float': np.float32,
        'int16': np.int16,
        'int32': np.int32,
        'uint16': np.uint16,
        'uint8': np.uint8
    }
    dtype = dtype_map.get(data_type)
    if dtype is None:
        raise ValueError(f"Type de données non pris en charge : {data_type}")
    
    # 📘 Ordre des octets (endianness)
    byte_order = header.get('imagedata byte order', 'LITTLEENDIAN').lower()
    endianess = '<' if 'little' in byte_order else '>'
    
    # 📘 Vérifie la taille réelle du fichier .img
    img_size = os.path.getsize(img_path)
    expected_size = np.prod(shape) * np.dtype(dtype).itemsize
    
    if img_size != expected_size:
        raise ValueError(f"La taille du fichier img ({img_size} octets) ne correspond pas à la taille attendue ({expected_size} octets).")
    
    # 📘 Lire les données binaires et les reformater
    with open(img_path, 'rb') as f:
        data = np.fromfile(f, dtype=endianess + np.dtype(dtype).char)
    
    image =  data.reshape(shape[::-1]) 
    
    # 📘 Rescale l'image si nécessaire
    rescale_slope = float(header.get('data rescale slope', 1))
    rescale_offset = float(header.get('data rescale offset', 0))
    image = image * rescale_slope + rescale_offset
    
    return image

def mse(y_true, y_pred):
    """
    Calcule la Mean Squared Error (MSE) entre deux tableaux.
    Équivalent à sklearn.metrics.mean_squared_error.
    """
    y_true = np.asarray(y_true)
    y_pred = np.asarray(y_pred)
    return np.mean((y_true - y_pred) ** 2)

def ssim(img1, img2, win_size=7, k1=0.01, k2=0.03, L=1.0):
    """
    Calcule l'SSIM entre deux images 2D (niveaux de gris).
    Équivalent à skimage.metrics.structural_similarity avec :
    - data_range=1.0 (images normalisées entre 0 et 1)
    - gaussian_weights=True (fenêtre gaussienne)
    - multichannel=False (1 canal)

    Args:
        img1, img2: Images 2D (numpy arrays) de même taille.
        win_size: Taille de la fenêtre de comparaison (doit être impair).
        k1, k2: Constantes pour stabiliser la division (valeurs typiques : 0.01, 0.03).
        L: Dynamique des pixels (1.0 si images dans [0,1], 255 si dans [0,255]).
    Returns:
        SSIM moyen sur l'image (float entre -1 et 1).
    """
    if img1.shape != img2.shape:
        raise ValueError("Les images doivent avoir la même taille.")
    if win_size % 2 == 0:
        raise ValueError("win_size doit être impair.")

    # Constantes
    C1 = (k1 * L) ** 2
    C2 = (k2 * L) ** 2

    # Fenêtre gaussienne
    window = np.ones((win_size, win_size)) / (win_size ** 2)  # Approximation (skimage utilise une gaussienne)
    window = window / np.sum(window)  # Normalisation

    # Pad les images pour éviter les bords
    pad = win_size // 2
    img1_pad = np.pad(img1, pad, mode='reflect')
    img2_pad = np.pad(img2, pad, mode='reflect')

    # Calcul des statistiques locales
    mu1 = np.zeros_like(img1, dtype=np.float64)
    mu2 = np.zeros_like(img1, dtype=np.float64)
    sigma1_sq = np.zeros_like(img1, dtype=np.float64)
    sigma2_sq = np.zeros_like(img1, dtype=np.float64)
    sigma12 = np.zeros_like(img1, dtype=np.float64)

    # Itère sur chaque pixel (optimisé avec des convolutions)
    for i in range(pad, img1_pad.shape[0] - pad):
        for j in range(pad, img1_pad.shape[1] - pad):
            patch1 = img1_pad[i-pad:i+pad+1, j-pad:j+pad+1]
            patch2 = img2_pad[i-pad:i+pad+1, j-pad:j+pad+1]

            mu1[i-pad, j-pad] = np.sum(patch1 * window)
            mu2[i-pad, j-pad] = np.sum(patch2 * window)
            sigma1_sq[i-pad, j-pad] = np.sum(window * (patch1 - mu1[i-pad, j-pad]) ** 2)
            sigma2_sq[i-pad, j-pad] = np.sum(window * (patch2 - mu2[i-pad, j-pad]) ** 2)
            sigma12[i-pad, j-pad] = np.sum(window * (patch1 - mu1[i-pad, j-pad]) * (patch2 - mu2[i-pad, j-pad]))

    # SSIM locale
    ssim_map = ((2 * mu1 * mu2 + C1) * (2 * sigma12 + C2)) / (
        (mu1**2 + mu2**2 + C1) * (sigma1_sq + sigma2_sq + C2)
    )

    return np.mean(ssim_map)

def calculate_memory_requirement(SMatrix, y):
    """
    Calcule la mémoire requise (en Go) pour :
    - SMatrix : Matrice (np.ndarray, CuPy CSR, SparseSMatrix_CSR ou SparseSMatrix_SELL)
    - y : vecteur (NumPy ou CuPy, float32)

    Args:
        SMatrix: Matrix object (np.ndarray, cpsparse.csr_matrix, SparseSMatrix_CSR, or SparseSMatrix_SELL)
        y: Vector (float32)
    """
    total_bytes = 0

    # --- 1. Memory for SMatrix ---
    
    # 1.1. Custom Sparse Matrix (SELL/CSR)
    if isinstance(SMatrix, (SparseSMatrix_SELL, SparseSMatrix_CSR)):
        # We rely on the getMatrixSize method, which we fixed to track all host/GPU bytes.
        # This is the most reliable way to estimate memory for custom GPU-backed structures.
        try:
            matrix_size_gb = SMatrix.getMatrixSize()
            if isinstance(matrix_size_gb, dict) and 'error' in matrix_size_gb:
                raise ValueError(f"SMatrix allocation error: {matrix_size_gb['error']}")
            
            # Convert GB back to bytes (1 GB = 1024^3 bytes)
            size_SMatrix = matrix_size_gb * (1024 ** 3)
            total_bytes += size_SMatrix
            print(f"SMatrix (Custom Sparse) size: {matrix_size_gb:.3f} GB")

        except AttributeError:
            raise AttributeError("Custom Sparse Matrix must implement the getMatrixSize() method.")
    
    # 1.2. NumPy Dense Array (Standard)
    elif isinstance(SMatrix, np.ndarray):
        # Dense NumPy array (float32)
        size_SMatrix = SMatrix.nbytes
        total_bytes += size_SMatrix
        print(f"SMatrix (NumPy Dense) size: {size_SMatrix / (1024 ** 3):.3f} GB")

    # 1.3. CuPy CSR Matrix (Standard Sparse CuPy)
    # Note: Requires CuPy to be imported, which is usually done outside this function.
    # Assuming 'cpsparse.csr_matrix' is available in the environment if this path is taken.
    elif 'cupy.sparse' in str(type(SMatrix)): # Using string check for type safety outside CuPy context
        # CuPy CSR matrix structure: data (float32), indices (int32), indptr (int32)
        nnz = SMatrix.nnz
        num_rows = SMatrix.shape[0]
        size_data = nnz * 4        # float32 = 4 bytes
        size_indices = nnz * 4     # int32 = 4 bytes
        size_indptr = (num_rows + 1) * 4 # int32 = 4 bytes
        size_SMatrix = size_data + size_indices + size_indptr
        total_bytes += size_SMatrix
        print(f"SMatrix (CuPy CSR) size: {size_SMatrix / (1024 ** 3):.3f} GB")

    else:
        raise ValueError("SMatrix must be a np.ndarray, cpsparse.csr_matrix, or a custom SparseSMatrix object (CSR/SELL).")

    # --- 2. Memory for Vector y ---
    
    # Check if y is a CuPy array or NumPy array (assuming float32 based on docstring)
    if hasattr(y, 'nbytes'):
        size_y = y.nbytes
        total_bytes += size_y
        print(f"Vector y size: {size_y / (1024 ** 3):.3f} GB")
    else:
        # Fallback if object doesn't expose nbytes (e.g., custom buffer), but usually array objects do.
        raise ValueError("Vector y must be an array type exposing the .nbytes attribute.")


    # --- 3. Final Result ---
    return total_bytes / (1024 ** 3)

def check_gpu_memory(device_index, required_memory, show_logs=True):
    """Check if enough memory is available on the specified GPU."""
    free_memory, _ = torch.cuda.mem_get_info(f"cuda:{device_index}")
    free_memory_gb = free_memory / 1024**3
    if show_logs:
        print(f"Free memory on GPU {device_index}: {free_memory_gb:.2f} GB, Required memory: {required_memory:.2f} GB")
    return free_memory_gb >= required_memory

@njit(parallel=True)
def _forward_projection(SMatrix, theta_p, q_p):
    t_dim, z_dim, x_dim, i_dim = SMatrix.shape
    for _t in prange(t_dim):
        for _n in range(i_dim):
            total = 0.0
            for _z in range(z_dim):
                for _x in range(x_dim):
                    total += SMatrix[_t, _z, _x, _n] * theta_p[_z, _x]
            q_p[_t, _n] = total

@njit(parallel=True)
def _backward_projection(SMatrix, e_p, c_p):
    t_dim, z_dim, x_dim, n_dim = SMatrix.shape
    for _z in prange(z_dim):
        for _x in range(x_dim):
            total = 0.0
            for _t in range(t_dim):
                for _n in range(n_dim):
                    total += SMatrix[_t, _z, _x, _n] * e_p[_t, _n]
            c_p[_z, _x] = total

def _build_adjacency_sparse(Z, X, device, corner=(0.5 - np.sqrt(2) / 4) / np.sqrt(2), face=0.5 - np.sqrt(2) / 4,dtype=torch.float32):
    rows, cols, weights = [], [], []
    for z in range(Z):
        for x in range(X):
            j = z * X + x
            for dz, dx in [(-1, -1), (-1, 0), (-1, 1),
                           (0, -1),           (0, 1),
                           (1, -1),   (1, 0), (1, 1)]:
                nz, nx = z + dz, x + dx
                if 0 <= nz < Z and 0 <= nx < X:
                    k = nz * X + nx
                    weight = corner if abs(dz) + abs(dx) == 2 else face
                    rows.append(j)
                    cols.append(k)
                    weights.append(weight)
    index = torch.tensor([rows, cols], dtype=torch.long, device=device)
    values = torch.tensor(weights, dtype=dtype, device=device)
    index, values = coalesce(index, values, m=Z*X, n=Z*X)
    return index, values

def power_method(P, PT, data, Z, X, n_it=10):
    x = torch.randn(Z * X, device=data.device)
    x = x / torch.norm(x)
    for _ in range(n_it):
        Ax = P(x)
        ATax = PT(Ax)
        x = ATax / torch.norm(ATax)
    ATax = PT(P(x))
    return torch.sqrt(torch.dot(x, ATax))

def proj_l2(p, alpha):
    if alpha <= 0:
        return torch.zeros_like(p)
    norm = torch.sqrt(torch.sum(p**2, dim=0, keepdim=True) + 1e-12)
    return p * torch.min(norm, torch.tensor(alpha, device=p.device)) / (norm + 1e-12)

def gradient(x):
    grad_x = torch.zeros_like(x)
    grad_y = torch.zeros_like(x)
    grad_x[:, :-1] = x[:, 1:] - x[:, :-1]  # Gradient horizontal
    grad_y[:-1, :] = x[1:, :] - x[:-1, :]   # Gradient vertical
    return torch.stack((grad_x, grad_y), dim=0)

def div(x):
    if x.dim() == 3:
        x = x.unsqueeze(0)  # Ajoute une dimension batch si nécessaire

    gx = x[:, 0, :, :]  # Gradient horizontal (shape: [1, H, W] ou [H, W])
    gy = x[:, 1, :, :]  # Gradient vertical   (shape: [1, H, W] ou [H, W])

    # Divergence du gradient horizontal (gx)
    div_x = torch.zeros_like(gx)
    div_x[:, :, 1:] += gx[:, :, :-1]  # Contribution positive (gauche)
    div_x[:, :, :-1] -= gx[:, :, :-1] # Contribution négative (droite)

    # Divergence du gradient vertical (gy)
    div_y = torch.zeros_like(gy)
    div_y[:, 1:, :] += gy[:, :-1, :]  # Contribution positive (haut)
    div_y[:, :-1, :] -= gy[:, :-1, :] # Contribution négative (bas)

    return -(div_x + div_y)

def norm2sq(x):
    return torch.sum(x**2)

def norm1(x):
    return torch.sum(torch.abs(x))

def KL_divergence(Ax, y):
    return torch.sum(Ax - y * torch.log(Ax + 1e-10))

def gradient_KL(Ax, y):
    return 1 - y / (Ax + 1e-10)

def prox_F_star(y, sigma, a):
    return 0.5 * (y - torch.sqrt(y**2 + 4 * sigma * a))

def prox_G(x, tau, K):
    return torch.clamp(x - tau * K, min=0)

def filter_radon(f, N, filter_type, Fc):
    """
    Implémente les filtres pour la rétroprojection filtrée (iRadon).
    Inspirée de la fonction MATLAB FilterRadon de Mamouna Bocoum.

    Paramètres :
    ------------
    f : np.ndarray
        Vecteur des fréquences (ex: f_t ou f_z).
    N : int
        Taille du filtre (longueur de f).
    filter_type : str
        Type de filtre : 'ram-lak', 'shepp-logan', 'cosine', 'hamming', 'hann'.
    Fc : float
        Fréquence de coupure.

    Retourne :
    -----------
    FILTER : np.ndarray
        Filtre appliqué aux fréquences.
    """
    FILTER = np.abs(f)

    if filter_type == 'ram-lak':
        pass  # FILTER = |f| (déjà calculé)
    elif filter_type == 'shepp-logan':
        # Évite la division par zéro
        with np.errstate(divide='ignore', invalid='ignore'):
            FILTER = FILTER * (np.sinc(2 * f / (2 * Fc)))  # sin(2πf/(2Fc))/(2πf/(4Fc)) = sinc(2f/(2Fc))
        FILTER[np.isnan(FILTER)] = 1.0  # Pour f=0
    elif filter_type == 'cosine':
        FILTER = FILTER * np.cos(2 * np.pi * f / (4 * Fc))
    elif filter_type == 'hamming':
        FILTER = FILTER * (0.54 + 0.46 * np.cos(2 * np.pi * f / Fc))
    elif filter_type == 'hann':
        FILTER = FILTER * (1 + np.cos(2 * np.pi * f / (4 * Fc))) / 2
    else:
        raise ValueError(f"Type de filtre inconnu : {filter_type}")

    # Coupure des fréquences au-delà de Fc
    FILTER[np.abs(f) > Fc] = 0
    # Atténuation exponentielle (optionnelle, comme dans le code MATLAB)
    FILTER = FILTER * np.exp(-2 * (np.abs(f) / Fc)**10)

    return FILTER

def compute_TV_cpu(x, Z, X, isotropic=False):
    """
    Compute total variation of x (1D flattened of shape Z*X).
    isotropic=False -> anisotropic (sum |dx| + |dy|)
    isotropic=True -> isotropic sqrt(dx^2 + dy^2)
    """
    x2d = x.reshape(Z, X)
    dx = np.diff(x2d, axis=1)
    dy = np.diff(x2d, axis=0)
    if isotropic:
        # pad to original size for consistent measure (we only need sum of norms)
        mags = np.sqrt(dx**2 + dy**2)
        return float(np.sum(mags))
    else:
        return float(np.sum(np.abs(dx)) + np.sum(np.abs(dy)))

def get_apodization_vector_gpu(matrix_sparse_obj):
    """
    Génère un vecteur de fenêtrage 2D (Hanning) pour l'apodisation 
    de la matrice système A et le transfère sur le GPU.
    Ce vecteur doit être multiplié par les colonnes de A (pixels Z*X).
    """
    Z = matrix_sparse_obj.Z
    X = matrix_sparse_obj.X
    
    # 1. Génération des fenêtres 1D sur l'axe X et Z
    # Forte apodisation latérale (X) pour cibler l'artefact de bordure.
    fenetre_x = hann(X).astype(np.float32)
    
    # Fenêtre uniforme en profondeur (Z), car l'artefact est surtout latéral.
    fenetre_z = np.ones(Z, dtype=np.float32)
    
    # 2. Création de la matrice de fenêtre 2D (Z, X)
    fenetre_2d = np.outer(fenetre_z, fenetre_x)
    
    # 3. Vectorisation (Z*X)
    fenetre_vectorisee = fenetre_2d.flatten()
    
    # 4. Transfert sur GPU (mémoire contiguë)
    fenetre_gpu = drv.mem_alloc(fenetre_vectorisee.nbytes)
    drv.memcpy_htod(fenetre_gpu, fenetre_vectorisee)
    
    print(f"✅ Vecteur de fenêtrage (Z*X={Z*X}) généré et transféré sur GPU.")
    
    return fenetre_gpu

def _call_axpby(axpby_kernel, out_ptr, x_ptr, y_ptr, a, b, N, stream, block):
    grid = ((int(N) + block - 1) // block, 1, 1)
    axpby_kernel(out_ptr, x_ptr, y_ptr,
                    np.float32(a), np.float32(b),
                    np.int32(N),
                    block=(block, 1, 1), grid=grid, stream=stream)

def _call_minus_axpy(minus_kernel, out_ptr, z_ptr, a, N, stream, block):
    grid = ((int(N) + block - 1) // block, 1, 1)
    minus_kernel(out_ptr, z_ptr, np.float32(a), np.int32(N),
                    block=(block, 1, 1), grid=grid, stream=stream)

def power_method_estimate_L__SELL(SMatrix, stream, n_it=20, block_size=256):
    """Estimate ||A||^2 using power method (uses your projection/backprojection kernels)."""
    TN = int(SMatrix.N * SMatrix.T)
    ZX = int(SMatrix.Z * SMatrix.X)
    proj = SMatrix.sparse_mod.get_function("projection_kernel__SELL")
    back = SMatrix.sparse_mod.get_function("backprojection_kernel__SELL")
    TN_i = np.int32(TN)
    ZX_i = np.int32(ZX)
    slice_h = np.int32(SMatrix.slice_height)
    grid_rows = ((TN + block_size - 1) // block_size, 1, 1)
    block_1D = (block_size, 1, 1)

    dtype = np.float32
    x_host = np.random.randn(ZX).astype(dtype)
    x_host /= np.linalg.norm(x_host) + 1e-12
    x_gpu = drv.mem_alloc(x_host.nbytes)
    drv.memcpy_htod_async(x_gpu, x_host, stream)
    q_gpu = drv.mem_alloc(TN * np.dtype(dtype).itemsize)
    ATq_gpu = drv.mem_alloc(ZX * np.dtype(dtype).itemsize)
    ATq_host = np.empty(ZX, dtype=dtype)

    for _ in range(n_it):
        proj(q_gpu, SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
                x_gpu, TN_i, slice_h, block=block_1D, grid=grid_rows, stream=stream)
        drv.memset_d32_async(ATq_gpu, 0, ZX, stream)
        back(SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
                q_gpu, ATq_gpu, TN_i, slice_h, block=block_1D, grid=grid_rows, stream=stream)
        stream.synchronize()
        drv.memcpy_dtoh(ATq_host, ATq_gpu)
        norm = np.linalg.norm(ATq_host)
        if norm < 1e-12:
            break
        x_host = ATq_host / norm
        drv.memcpy_htod_async(x_gpu, x_host, stream)
    # final Rayleigh quotient
    proj(q_gpu, SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
            x_gpu, TN_i, slice_h, block=block_1D, grid=grid_rows, stream=stream)
    drv.memset_d32_async(ATq_gpu, 0, ZX, stream)
    back(SMatrix.sell_values_gpu, SMatrix.sell_colinds_gpu, SMatrix.slice_ptr_gpu, SMatrix.slice_len_gpu,
            q_gpu, ATq_gpu, TN_i, slice_h, block=block_1D, grid=grid_rows, stream=stream)
    stream.synchronize()
    drv.memcpy_dtoh(ATq_host, ATq_gpu)
    L_sq = float(np.dot(x_host, ATq_host))
    for g in (x_gpu, q_gpu, ATq_gpu):
        try:
            g.free()
        except:
            pass
    return max(L_sq, 1e-6)

def get_phase_deterministic(profile):
    """
    Détermine la phase en se basant sur la valeur initiale (0 ou 1) et l'état
    de décalage (is_shifted) de la séquence binaire.
    
    ATTENTION: Cette fonction est conservée mais la logique est souvent simplifiée
    en pratique si les labels garantissent les phases 0, pi/2, pi, 3pi/2.
    """
    runs = [(k, sum(1 for _ in g)) for k, g in groupby(profile)]
    if not runs: return 0.0
    
    nominal_half_period = max([r[1] for r in runs]) 
    if nominal_half_period == 0: return 0.0

    first_val = runs[0][0] # 0 ou 1
    first_len = runs[0][1] 
    # Détection de cycle 50%
    is_shifted = (0.3 < first_len / nominal_half_period < 0.7) 
    
    # --- LOGIQUE DE MAPPAGE DE PHASE SIMPLIFIÉE (idx 1 à 4) ---
    
    if first_val == 0: 
        if is_shifted:
            idx = 3 # C1/C3 décalé (phi_1 ou phi_3)
        else:
            idx = 4 # C2/C4 non décalé
    else: # first_val == 1
        if is_shifted:
            idx = 1 # C1/C3 décalé (phi_1 ou phi_3)
        else:
            idx = 2 # C2/C4 non décalé

    # On utilise les phases de quadrature 0, pi/2, pi, 3pi/2 
    if idx == 1:
        phase = 0
    elif idx == 2 :
        phase = np.pi/2
    elif idx == 3 :
        phase = np.pi
    elif idx == 4 :
        phase = 3*np.pi/2
            
    return phase

def add_sincos_cpu(R, decimation, theta):
    decimation = np.asarray(decimation)
    theta = np.asarray(theta)

    ScanParam = np.stack([decimation, theta], axis=1)
    uniq, ia, ib = np.unique(ScanParam, axis=0, return_index=True, return_inverse=True)

    theta_u = uniq[:,1]
    decim_u = uniq[:,0]

    theta0 = np.unique(theta_u)
    N0 = len(theta0)

    Rg = np.asarray(R)
    Nz = Rg.shape[0]
    Nk = N0 + (Rg.shape[1] - N0)//4

    Iout = np.zeros((Nz, Nk), dtype=np.complex64)
    # fx = 0 (onde plane)
    Iout[:, :N0] = Rg[:, :N0]

    k = N0
    for i in range(N0, len(ia)):
        idx = np.where(ib == i)[0]
        h1, h2, h3, h4 = Rg[:, idx].T
        Iout[:, k] = ((h1 - h2) - 1j*(h3 - h4)) / 2
        k += 1

    return Iout, theta_u, decim_u

def fourierz_gpu(z, X):
    dz = float(z[1] - z[0])
    Nz = X.shape[0]

    return cp.fft.fftshift(
        cp.fft.fft(
            cp.fft.ifftshift(X, axes=0),
            axis=0
        ),
        axes=0
    ) * (Nz * dz)

def ifourierz_gpu(z, X):
    dz = float(z[1] - z[0])
    Nz = X.shape[0]

    return cp.fft.ifftshift(
        cp.fft.ifft(
            cp.fft.fftshift(X, axes=0),
            axis=0
        ),
        axes=0
    ) * (1 / dz)

def ifourierx_gpu(F_fx_z, dx):
    """
    Inverse Fourier along X (axis=1), Matlab-compatible
    F_fx_z : (Nz, Nx) complex cupy array
    dx : scalar (spacing in x)
    """

    return (
        cp.fft.ifftshift(
            cp.fft.ifft(
                cp.fft.fftshift(F_fx_z, axes=1),
                axis=1
            ),
            axes=1
        ) * (1.0 / dx)
    )

def EvalDelayLawOS_center(X_m, theta, DelayLAWS, ActiveLIST, c):
    """
    Retourne le centre de rotation C pour chaque angle
    X_m : positions des éléments de la sonde
    DelayLAWS : delays en secondes (chaque colonne = angle, chaque ligne = élément)
    ActiveLIST : masque des éléments actifs (1 = actif)
    c : vitesse du son
    """
    Nangle = DelayLAWS.shape[1]
    C = np.zeros((Nangle, 2))
    
    ct = DelayLAWS * c  # convert seconds to distance

    for i in range(Nangle):
        active_idx = np.where(ActiveLIST[:, i] == 1)[0]
        if len(active_idx) == 0:
            continue


        angle_i = np.round(theta[i], 5)
        # unit vector orthogonal to wavefront
        u = np.array([np.sin(angle_i), np.cos(angle_i)])

        # initial positions X0, Z0
        X0 = X_m - u[0] * ct[:, i]
        Z0 = 0 - u[1] * ct[:, i]

        if Z0[-1] - Z0[0] != 0:
            C[i, 0] = (Z0[-1]*X0[0] - Z0[0]*X0[-1]) / (Z0[-1] - Z0[0])
            C[i, 1] = 0

    return C

def rotate_theta_gpu(X, Z, Iin, theta, C):
    """
    GPU equivalent of RotateTheta.m
    X, Z, Iin : cupy arrays (Nz, Nx)
    theta : scalar (float)
    C : (2,) array-like
    """

    # --- Translation ---
    X_rel = X - C[0]
    Z_rel = Z - C[1]

    c = cp.cos(theta)
    s = cp.sin(theta)

    # --- Rotation (Matlab convention) ---
    Xout = c * X_rel + s * Z_rel
    Zout = -s * X_rel + c * Z_rel

    # Back to original frame
    Xout += C[0]
    Zout += C[1]

    # --- Conversion coordonnées -> indices ---
    # Grille régulière supposée
    dx = X[0, 1] - X[0, 0]
    dz = Z[1, 0] - Z[0, 0]

    x0 = X[0, 0]
    z0 = Z[0, 0]

    ix = (Xout - x0) / dx
    iz = (Zout - z0) / dz

    # --- Interpolation bilinéaire GPU ---
    # map_coordinates attend (ndim, Npoints)
    coords = cp.stack([iz.ravel(), ix.ravel()])

    Iout = map_coordinates(
        Iin,
        coords,
        order=1,          # bilinear
        mode='constant',
        cval=0.0
    )

    return Iout.reshape(Iin.shape)

def filter_radon_gpu(fz, Fc):
    FILTER = cp.abs(fz)
    FILTER = cp.where(cp.abs(fz) > Fc, 0, FILTER)
    FILTER *= cp.exp(-2 * cp.abs(fz / Fc)**10)
    return FILTER

